from typing import Optional

from .base import KittyCadBaseModel


class EmailAuthenticationForm(KittyCadBaseModel):
    """The body of the form for email authentication."""

    callback_url: Optional[str] = None

    email: str



from typing import Optional

from .base import KittyCadBaseModel


class UpdateShortlinkRequest(KittyCadBaseModel):
    """Request to update a shortlink."""

    password: Optional[str] = None

    restrict_to_org: bool



from enum import Enum


class ModelingAppShareLinks(str, Enum):
    """The options for sharable links through the modeling app."""  # noqa: E501

    """# Public."""  # noqa: E501

    PUBLIC = "public"

    """# Password protected."""  # noqa: E501

    PASSWORD_PROTECTED = "password_protected"

    """# Organization only. Links can be made only available to members of the organization."""  # noqa: E501

    ORGANIZATION_ONLY = "organization_only"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class ApiCallStatus(str, Enum):
    """The status of an async API call."""  # noqa: E501

    """# The async API call is queued."""  # noqa: E501

    QUEUED = "queued"

    """# The async API call was uploaded to be converted."""  # noqa: E501

    UPLOADED = "uploaded"

    """# The async API call is in progress."""  # noqa: E501

    IN_PROGRESS = "in_progress"

    """# The async API call has completed."""  # noqa: E501

    COMPLETED = "completed"

    """# The async API call has failed."""  # noqa: E501

    FAILED = "failed"

    def __str__(self) -> str:
        return str(self.value)



import json
import os
import time
import uuid

import pytest
from websockets.exceptions import ConnectionClosedError

from kittycad import KittyCAD
from kittycad.models import (
    ApiCallStatus,
    Axis,
    AxisDirectionPair,
    ConversionParams,
    CreatedAtSortMode,
    Direction,
    FileCenterOfMass,
    FileConversion,
    FileExportFormat,
    FileImportFormat,
    FileMass,
    FileVolume,
    ImageFormat,
    ImportFile,
    InputFormat3d,
    ModelingCmd,
    ModelingCmdId,
    OutputFormat3d,
    Pong,
    PostEffectType,
    System,
    TextToCadCreateBody,
    UnitDensity,
    UnitLength,
    UnitMass,
    UnitVolume,
    User,
    WebSocketRequest,
    WebSocketResponse,
)
from kittycad.models.input_format3d import OptionObj, OptionStl
from kittycad.models.modeling_cmd import (
    OptionDefaultCameraFocusOn,
    OptionImportFiles,
    OptionStartPath,
    OptionTakeSnapshot,
)
from kittycad.models.output_format3d import OptionObj as OutputOptionObj
from kittycad.models.web_socket_request import OptionModelingCmdReq
from kittycad.types import Unset


def _poll_for_completion(
    client, fc: FileConversion, timeout_seconds: int = 60
) -> FileConversion:
    """Poll for file conversion completion.

    Args:
        client: KittyCAD client (sync or async)
        fc: Initial FileConversion object
        timeout_seconds: Maximum time to wait for completion

    Returns:
        Completed FileConversion object
    """
    import time

    start_time = time.time()

    # Handle both completed and in-progress cases
    if fc.status == ApiCallStatus.COMPLETED:
        # Already completed, no need to poll
        return fc

    # Need to poll for completion
    current_status: ApiCallStatus = fc.status
    body = fc

    while (
        current_status == ApiCallStatus.IN_PROGRESS
        or current_status == ApiCallStatus.QUEUED
        or current_status == ApiCallStatus.UPLOADED
    ) and time.time() - start_time < timeout_seconds:
        time.sleep(2)  # Wait 2 seconds between polls
        result_status = client.api_calls.get_async_operation(id=fc.id)

        # Handle response from get_async_operation
        if isinstance(result_status, dict):
            status_str = result_status.get("status", "unknown")
            print(f"FileConversion status: {status_str}")
            # Convert string status to ApiCallStatus
            if status_str == "completed":
                current_status = ApiCallStatus.COMPLETED
            elif status_str == "failed":
                current_status = ApiCallStatus.FAILED
            elif status_str == "in_progress":
                current_status = ApiCallStatus.IN_PROGRESS
            elif status_str == "queued":
                current_status = ApiCallStatus.QUEUED
            elif status_str == "uploaded":
                current_status = ApiCallStatus.UPLOADED
            else:
                current_status = ApiCallStatus.FAILED  # Default for unknown

            if current_status in [ApiCallStatus.COMPLETED, ApiCallStatus.FAILED]:
                break
        else:
            if hasattr(result_status, "status"):
                current_status = result_status.status
                body = result_status
                print(f"FileConversion status: {current_status}")

    # If we exited the loop due to completion, get the final FileConversion object
    if current_status == ApiCallStatus.COMPLETED:
        # Get the final state of the FileConversion object
        final_result = client.api_calls.get_async_operation(id=fc.id)
        if not isinstance(final_result, dict):
            body = final_result

    return body


async def _poll_for_completion_async(
    client, fc: FileConversion, timeout_seconds: int = 60
) -> FileConversion:
    """Async version of poll for file conversion completion.

    Args:
        client: AsyncKittyCAD client
        fc: Initial FileConversion object
        timeout_seconds: Maximum time to wait for completion

    Returns:
        Completed FileConversion object
    """
    import time

    start_time = time.time()

    # Handle both completed and in-progress cases
    if fc.status == ApiCallStatus.COMPLETED:
        # Already completed, no need to poll
        return fc

    # Need to poll for completion
    current_status: ApiCallStatus = fc.status
    body = fc

    while (
        current_status == ApiCallStatus.IN_PROGRESS
        or current_status == ApiCallStatus.QUEUED
        or current_status == ApiCallStatus.UPLOADED
    ) and time.time() - start_time < timeout_seconds:
        time.sleep(2)  # Wait 2 seconds between polls
        result_status = await client.api_calls.get_async_operation(id=fc.id)

        # Handle response from get_async_operation
        if isinstance(result_status, dict):
            status_str = result_status.get("status", "unknown")
            print(f"FileConversion status: {status_str}")
            # Convert string status to ApiCallStatus
            if status_str == "completed":
                current_status = ApiCallStatus.COMPLETED
            elif status_str == "failed":
                current_status = ApiCallStatus.FAILED
            elif status_str == "in_progress":
                current_status = ApiCallStatus.IN_PROGRESS
            elif status_str == "queued":
                current_status = ApiCallStatus.QUEUED
            elif status_str == "uploaded":
                current_status = ApiCallStatus.UPLOADED
            else:
                current_status = ApiCallStatus.FAILED  # Default for unknown

            if current_status in [ApiCallStatus.COMPLETED, ApiCallStatus.FAILED]:
                break
        else:
            if hasattr(result_status, "status"):
                current_status = result_status.status
                body = result_status
                print(f"FileConversion status: {current_status}")

    # If we exited the loop due to completion, get the final FileConversion object
    if current_status == ApiCallStatus.COMPLETED:
        # Get the final state of the FileConversion object
        final_result = await client.api_calls.get_async_operation(id=fc.id)
        if not isinstance(final_result, dict):
            body = final_result

    return body


def test_get_session():
    # Create our client
    client = KittyCAD()

    # Get the session using modern pattern
    session = client.users.get_user_self()

    assert isinstance(session, User)

    print(f"Session: {session}")


@pytest.mark.asyncio
async def test_get_api_tokens_async():
    # Create our client
    client = KittyCAD()

    # List API tokens using modern pattern
    fc = client.api_tokens.list_api_tokens_for_user(
        sort_by=CreatedAtSortMode.CREATED_AT_ASCENDING
    )

    # Now returns SyncPageIterator instead of ApiTokenResultsPage
    from kittycad.pagination import SyncPageIterator

    assert isinstance(fc, SyncPageIterator)

    print(f"fc: {fc}")


@pytest.mark.asyncio
async def test_get_session_async():
    from kittycad import AsyncKittyCAD

    # Create our async client
    client = AsyncKittyCAD()

    # Get the session using new async pattern
    session = await client.users.get_user_self()

    assert isinstance(session, User)

    print(f"Session: {session}")


def test_ping():
    # Create our client
    client = KittyCAD()

    # Get the message using modern pattern (no .api layer)
    message = client.meta.ping()

    assert isinstance(message, Pong)

    print(f"Message: {message}")


@pytest.mark.asyncio
async def test_ping_async():
    from kittycad import AsyncKittyCAD

    # Create our async client
    client = AsyncKittyCAD()

    # Get the message using new async pattern (no .api layer)
    message = await client.meta.ping()

    assert isinstance(message, Pong)

    print(f"Message: {message}")


def test_file_convert_stl():
    # Create our client
    client = KittyCAD()

    dir_path = os.path.dirname(os.path.realpath(__file__))
    file = open(os.path.join(dir_path, "../../assets/testing.stl"), "rb")
    content = file.read()
    file.close()

    # Get the file conversion using modern pattern
    fc = client.file.create_file_conversion(
        body=content,
        src_format=FileImportFormat.STL,
        output_format=FileExportFormat.OBJ,
    )

    assert isinstance(fc, FileConversion)

    print(f"FileConversion: {fc}")

    assert fc.id is not None
    # File conversion might complete immediately or be uploaded for async processing
    assert fc.status in [ApiCallStatus.UPLOADED, ApiCallStatus.COMPLETED]

    print(f"FileConversion: {fc}")

    assert not isinstance(fc.outputs, Unset)
    assert fc.outputs is not None

    # Make sure the bytes are not empty.
    for key, value in fc.outputs.items():
        assert len(value) > 0


@pytest.mark.asyncio
async def test_file_convert_stl_async():
    from kittycad import AsyncKittyCAD

    # Create our async client
    client = AsyncKittyCAD()

    dir_path = os.path.dirname(os.path.realpath(__file__))
    file = open(os.path.join(dir_path, "../../assets/testing.stl"), "rb")
    content = file.read()
    file.close()

    # Get the file conversion using new async pattern
    result = await client.file.create_file_conversion(
        body=content,
        src_format=FileImportFormat.STL,
        output_format=FileExportFormat.OBJ,
    )

    assert isinstance(result, FileConversion)

    fc: FileConversion = result

    print(f"FileConversion: {fc}")

    assert fc.id is not None
    # File conversion might complete immediately or be uploaded for async processing
    assert fc.status in [ApiCallStatus.UPLOADED, ApiCallStatus.COMPLETED]

    print(f"FileConversion: {fc}")

    assert not isinstance(fc.outputs, Unset)
    assert fc.outputs is not None

    # Make sure the bytes are not empty.
    for key, value in fc.outputs.items():
        assert len(value) > 0


@pytest.mark.asyncio
async def test_file_convert_obj_async():
    from kittycad import AsyncKittyCAD

    # Create our async client
    client = AsyncKittyCAD()

    dir_path = os.path.dirname(os.path.realpath(__file__))
    file = open(os.path.join(dir_path, "../../assets/ORIGINALVOXEL-3.obj"), "rb")
    content = file.read()
    file.close()

    # Get the file conversion using new async pattern
    result = await client.file.create_file_conversion(
        body=content,
        src_format=FileImportFormat.OBJ,
        output_format=FileExportFormat.STL,
    )

    assert isinstance(result, FileConversion)

    fc: FileConversion = result

    print(f"FileConversion: {fc}")

    assert fc.id is not None
    # File conversion might complete immediately or be uploaded for async processing
    assert fc.status in [ApiCallStatus.UPLOADED, ApiCallStatus.COMPLETED]

    print(f"FileConversion: {fc}")

    assert not isinstance(fc.outputs, Unset)
    assert fc.outputs is not None

    # Make sure the bytes are not empty.
    for key, value in fc.outputs.items():
        assert len(value) > 0


def test_file_conversion_options_stl():
    # Create our client
    client = KittyCAD()

    dir_path = os.path.dirname(os.path.realpath(__file__))
    file_path = os.path.join(dir_path, "../../assets/testing.stl")

    # Test create_file_conversion_options with the same input file as create_file_conversion
    fc = client.file.create_file_conversion_options(
        body=ConversionParams(
            src_format=InputFormat3d(
                OptionStl(
                    coords=System(
                        forward=AxisDirectionPair(
                            axis=Axis.Y,
                            direction=Direction.NEGATIVE,
                        ),
                        up=AxisDirectionPair(
                            axis=Axis.Z,
                            direction=Direction.POSITIVE,
                        ),
                    ),
                    units=UnitLength.MM,
                )
            ),
            output_format=OutputFormat3d(
                OutputOptionObj(
                    coords=System(
                        forward=AxisDirectionPair(
                            axis=Axis.Y,
                            direction=Direction.POSITIVE,
                        ),
                        up=AxisDirectionPair(
                            axis=Axis.Z,
                            direction=Direction.POSITIVE,
                        ),
                    ),
                    units=UnitLength.MM,
                )
            ),
        ),
        file_attachments={"input.stl": file_path},
    )

    assert isinstance(fc, FileConversion)

    print(f"FileConversion initial: {fc}")

    assert fc.id is not None
    # File conversion might complete immediately or be uploaded for async processing
    assert fc.status in [ApiCallStatus.UPLOADED, ApiCallStatus.COMPLETED]

    # Poll for completion (max 60 seconds)
    body = _poll_for_completion(client, fc, timeout_seconds=60)

    # Check final status
    assert body.status == ApiCallStatus.COMPLETED
    print(f"FileConversion completed: {body}")

    assert not isinstance(body.outputs, Unset)
    assert body.outputs is not None

    # Make sure the bytes are not empty.
    for key, value in body.outputs.items():
        assert len(value) > 0


@pytest.mark.asyncio
async def test_file_conversion_options_stl_async():
    from kittycad import AsyncKittyCAD

    # Create our async client
    client = AsyncKittyCAD()

    dir_path = os.path.dirname(os.path.realpath(__file__))
    file_path = os.path.join(dir_path, "../../assets/testing.stl")

    # Test async create_file_conversion_options with the same input file
    result = await client.file.create_file_conversion_options(
        body=ConversionParams(
            src_format=InputFormat3d(
                OptionStl(
                    coords=System(
                        forward=AxisDirectionPair(
                            axis=Axis.Y,
                            direction=Direction.NEGATIVE,
                        ),
                        up=AxisDirectionPair(
                            axis=Axis.Z,
                            direction=Direction.POSITIVE,
                        ),
                    ),
                    units=UnitLength.MM,
                )
            ),
            output_format=OutputFormat3d(
                OutputOptionObj(
                    coords=System(
                        forward=AxisDirectionPair(
                            axis=Axis.Y,
                            direction=Direction.POSITIVE,
                        ),
                        up=AxisDirectionPair(
                            axis=Axis.Z,
                            direction=Direction.POSITIVE,
                        ),
                    ),
                    units=UnitLength.MM,
                )
            ),
        ),
        file_attachments={"input.stl": file_path},
    )

    assert isinstance(result, FileConversion)

    fc: FileConversion = result

    print(f"FileConversion initial: {fc}")

    assert fc.id is not None
    # File conversion might complete immediately or be uploaded for async processing
    assert fc.status in [ApiCallStatus.UPLOADED, ApiCallStatus.COMPLETED]

    # Poll for completion (max 60 seconds)
    body = await _poll_for_completion_async(client, fc, timeout_seconds=60)

    # Check final status
    assert body.status == ApiCallStatus.COMPLETED
    print(f"FileConversion completed: {body}")

    assert not isinstance(body.outputs, Unset)
    assert body.outputs is not None

    # Make sure the bytes are not empty.
    for key, value in body.outputs.items():
        assert len(value) > 0


@pytest.mark.asyncio
async def test_file_conversion_options_obj_async():
    from kittycad import AsyncKittyCAD

    # Create our async client
    client = AsyncKittyCAD()

    dir_path = os.path.dirname(os.path.realpath(__file__))
    file_path = os.path.join(dir_path, "../../assets/ORIGINALVOXEL-3.obj")

    # Test async create_file_conversion_options with OBJ input
    result = await client.file.create_file_conversion_options(
        body=ConversionParams(
            src_format=InputFormat3d(
                OptionObj(
                    coords=System(
                        forward=AxisDirectionPair(
                            axis=Axis.Y,
                            direction=Direction.NEGATIVE,
                        ),
                        up=AxisDirectionPair(
                            axis=Axis.Z,
                            direction=Direction.POSITIVE,
                        ),
                    ),
                    units=UnitLength.M,
                )
            ),
            output_format=OutputFormat3d(
                OutputOptionObj(
                    coords=System(
                        forward=AxisDirectionPair(
                            axis=Axis.Y,
                            direction=Direction.POSITIVE,
                        ),
                        up=AxisDirectionPair(
                            axis=Axis.Z,
                            direction=Direction.POSITIVE,
                        ),
                    ),
                    units=UnitLength.MM,
                )
            ),
        ),
        file_attachments={"input.obj": file_path},
    )

    assert isinstance(result, FileConversion)

    fc: FileConversion = result

    print(f"FileConversion initial: {fc}")

    assert fc.id is not None
    # File conversion might complete immediately or be uploaded for async processing
    assert fc.status in [ApiCallStatus.UPLOADED, ApiCallStatus.COMPLETED]

    # Poll for completion (max 60 seconds)
    body = await _poll_for_completion_async(client, fc, timeout_seconds=60)

    # Check final status
    assert body.status == ApiCallStatus.COMPLETED
    print(f"FileConversion completed: {body}")

    assert not isinstance(body.outputs, Unset)
    assert body.outputs is not None

    # Make sure the bytes are not empty.
    for key, value in body.outputs.items():
        assert len(value) > 0


def test_file_mass():
    # Create our client
    client = KittyCAD()

    dir_path = os.path.dirname(os.path.realpath(__file__))
    file = open(os.path.join(dir_path, "../../assets/testing.obj"), "rb")
    content = file.read()
    file.close()

    # Get the file mass using modern pattern
    fm = client.file.create_file_mass(
        body=content,
        src_format=FileImportFormat.OBJ,
        material_density=1.0,
        material_density_unit=UnitDensity.KG_M3,
        output_unit=UnitMass.G,
    )

    assert isinstance(fm, FileMass)

    print(f"FileMass: {fm}")

    assert fm.id is not None
    assert fm.mass is not None

    assert fm.model_dump_json() is not None

    assert fm.status == ApiCallStatus.COMPLETED


def test_file_volume():
    # Create our client
    client = KittyCAD()

    dir_path = os.path.dirname(os.path.realpath(__file__))
    file = open(os.path.join(dir_path, "../../assets/testing.obj"), "rb")
    content = file.read()
    file.close()

    # Get the file volume using modern pattern
    fv = client.file.create_file_volume(
        body=content,
        src_format=FileImportFormat.OBJ,
        output_unit=UnitVolume.CM3,
    )

    assert isinstance(fv, FileVolume)

    print(f"FileVolume: {fv}")

    assert fv.id is not None
    assert fv.volume is not None

    assert fv.model_dump_json() is not None

    assert fv.status == ApiCallStatus.COMPLETED


def test_file_center_of_mass():
    # Create our client
    client = KittyCAD()

    dir_path = os.path.dirname(os.path.realpath(__file__))
    file = open(os.path.join(dir_path, "../../assets/testing.obj"), "rb")
    content = file.read()
    file.close()

    # Get the file center of mass using modern pattern
    fv = client.file.create_file_center_of_mass(
        body=content,
        src_format=FileImportFormat.OBJ,
        output_unit=UnitLength.CM,
    )

    assert isinstance(fv, FileCenterOfMass)

    print(f"FileCenterOfMass: {fv}")

    assert fv.id is not None
    assert fv.center_of_mass is not None

    assert fv.model_dump_json() is not None

    assert fv.status == ApiCallStatus.COMPLETED


def test_list_users():
    # Create our client
    client = KittyCAD()

    # List users using modern pattern
    response = client.users.list_users_extended(
        sort_by=CreatedAtSortMode.CREATED_AT_DESCENDING, limit=10
    )

    # Now returns SyncPageIterator instead of ExtendedUserResultsPage
    from kittycad.pagination import SyncPageIterator

    assert isinstance(response, SyncPageIterator)

    print(f"ExtendedUserResultsPage: {response}")


def test_ws_simple():
    # Create our client
    client = KittyCAD()

    # WebSocket uses direct pattern
    with client.modeling.modeling_commands_ws(
        fps=30,
        show_grid=False,
        post_effect=PostEffectType.NOEFFECT,
        unlocked_framerate=False,
        video_res_height=360,
        video_res_width=480,
        webrtc=False,
    ) as websocket:
        # Send a message.
        id = uuid.uuid4()
        req = WebSocketRequest(
            OptionModelingCmdReq(
                cmd=ModelingCmd(OptionStartPath()), cmd_id=ModelingCmdId(id)
            )
        )
        websocket.send(req)

        # Get the messages.
        while True:
            message = websocket.recv()
            print(json.dumps(message.model_dump_json()))
            break


def test_ws_import():
    max_retries = 3
    for attempt in range(1, max_retries + 1):
        try:
            # Create our client
            client = KittyCAD()

            # WebSocket uses direct pattern
            with client.modeling.modeling_commands_ws(
                fps=30,
                post_effect=PostEffectType.NOEFFECT,
                show_grid=False,
                unlocked_framerate=False,
                video_res_height=360,
                video_res_width=480,
                webrtc=False,
            ) as websocket:
                # read the content of the file
                dir_path = os.path.dirname(os.path.realpath(__file__))
                file_name = "ORIGINALVOXEL-3.obj"
                file_path = os.path.join(dir_path, "../..", "assets", file_name)
                with open(file_path, "rb") as file:
                    content = file.read()
                cmd_id = uuid.uuid4()
                ImportFile(data=content, path=file_name)
                # form the request
                req = WebSocketRequest(
                    OptionModelingCmdReq(
                        cmd=ModelingCmd(
                            OptionImportFiles(
                                files=[ImportFile(data=content, path=file_name)],
                                format=InputFormat3d(
                                    OptionObj(
                                        units=UnitLength.M,
                                        coords=System(
                                            forward=AxisDirectionPair(
                                                axis=Axis.Y,
                                                direction=Direction.NEGATIVE,
                                            ),
                                            up=AxisDirectionPair(
                                                axis=Axis.Z,
                                                direction=Direction.POSITIVE,
                                            ),
                                        ),
                                    )
                                ),
                            )
                        ),
                        cmd_id=ModelingCmdId(cmd_id),
                    )
                )
                # Import files request must be sent as binary, because the file contents might be binary.
                websocket.send_binary(req)

                # Get the success message.
                for message in websocket:
                    message_dict = message.model_dump()
                    if message_dict["success"] is not True:
                        raise Exception(message_dict)
                    elif message_dict["resp"]["type"] != "modeling":
                        continue
                    elif (
                        message_dict["resp"]["data"]["modeling_response"]["type"]
                        != "import_files"
                    ):
                        # We have a modeling command response.
                        # Make sure its the import files response.
                        raise Exception(message_dict)
                    else:
                        # Okay we have the import files response.
                        # Break since now we know it was a success.
                        object_id = str(
                            message_dict["resp"]["data"]["modeling_response"]["data"][
                                "object_id"
                            ]
                        )
                        break

                # Now we want to focus on the object.
                cmd_id = uuid.uuid4()
                # form the request
                req = WebSocketRequest(
                    OptionModelingCmdReq(
                        cmd=ModelingCmd(OptionDefaultCameraFocusOn(uuid=object_id)),
                        cmd_id=ModelingCmdId(cmd_id),
                    )
                )
                websocket.send(req)

                # Get the success message.
                for message in websocket:
                    message_dict = message.model_dump()
                    if message_dict["success"] is not True:
                        raise Exception(message_dict)
                    elif message_dict["resp"]["type"] != "modeling":
                        continue
                    elif message_dict["request_id"] == str(cmd_id):
                        # We got a success response for our cmd.
                        break
                    else:
                        raise Exception(message_dict)

                # Now we want to snapshot as a png.
                cmd_id = uuid.uuid4()
                # form the request
                req = WebSocketRequest(
                    OptionModelingCmdReq(
                        cmd=ModelingCmd(OptionTakeSnapshot(format=ImageFormat.PNG)),
                        cmd_id=ModelingCmdId(cmd_id),
                    )
                )
                websocket.send(req)

                # Get the success message.
                for message in websocket:
                    message_dict = message.model_dump()
                    if message_dict["success"] is not True:
                        raise Exception(message_dict)
                    elif message_dict["resp"]["type"] != "modeling":
                        continue
                    elif (
                        message_dict["resp"]["data"]["modeling_response"]["type"]
                        != "take_snapshot"
                    ):
                        # Make sure its the correct response.
                        raise Exception(message_dict)
                    else:
                        # Okay we have the snapshot response.
                        # Break since now we know it was a success.
                        png_contents = message_dict["resp"]["data"][
                            "modeling_response"
                        ]["data"]["contents"]
                        break

                # Save the contents to a file.
                png_path = os.path.join(dir_path, "../..", "assets", "snapshot.png")
                with open(png_path, "wb") as f:
                    f.write(png_contents)

                # Ensure the file is not empty.
                assert len(png_contents) > 0

                # Ensure the file exists.
                assert os.path.exists(png_path)

            # Exit the retry loop on success
            break

        except ConnectionClosedError:
            if attempt < max_retries:
                print(
                    f"ConnectionClosedError encountered on attempt {attempt}/{max_retries}. Retrying..."
                )
            else:
                # After max retries, re-raise the exception to fail the test
                print(
                    f"ConnectionClosedError encountered on attempt {attempt}/{max_retries}. No more retries left."
                )
                raise


def test_serialize_deserialize():
    json_str = """{"success":true,"request_id":"16a06065-6ca3-4a96-a042-d0bec6b161a6","resp":{"type":"modeling","data":{"modeling_response":{"type":"import_files","data":{"object_id":"f61ac02e-77bd-468f-858f-fd4141a26acd"}}}}}"""
    d = json.loads(json_str)
    print(d)
    message = WebSocketResponse(**d)
    model_dump = message.model_dump()
    print(model_dump)
    assert model_dump["success"] is True  # type: ignore
    assert model_dump["request_id"] == "16a06065-6ca3-4a96-a042-d0bec6b161a6"  # type: ignore
    assert model_dump["resp"]["type"] == "modeling"  # type: ignore
    assert model_dump["resp"]["data"]["modeling_response"]["type"] == "import_files"  # type: ignore
    assert (
        model_dump["resp"]["data"]["modeling_response"]["data"]["object_id"]
        == "f61ac02e-77bd-468f-858f-fd4141a26acd"
    )  # type: ignore


def test_deserialize_null_request_id():
    json_str = """{"success":true,"request_id":null,"resp":{"type":"modeling_session_data","data":{"session":{"api_call_id":"91f7fd17-8846-4593-97ff-6400a81b8cdd"}}}}"""
    d = json.loads(json_str)
    print(d)
    message = WebSocketResponse(**d)
    model_dump = message.model_dump()
    print(model_dump)
    assert model_dump["success"] is True  # type: ignore
    assert model_dump["success"] is True  # type: ignore
    assert model_dump["request_id"] is None  # type: ignore
    assert model_dump["resp"]["type"] == "modeling_session_data"  # type: ignore
    assert (
        model_dump["resp"]["data"]["session"]["api_call_id"]
        == "91f7fd17-8846-4593-97ff-6400a81b8cdd"
    )  # type: ignore


def test_text_to_cad():
    # Test the modern client.api pattern
    client = KittyCAD()

    # Modern way: client.ml.create_text_to_cad()
    result = client.ml.create_text_to_cad(
        output_format=FileExportFormat.STEP,
        body=TextToCadCreateBody(
            prompt="a 2x4 lego",
        ),
    )
    print(f"Modern result: {result}")

    # Poll the api until the status is completed.
    # Timeout after some seconds.
    start_time = time.time()
    body = result
    while (
        body.status == ApiCallStatus.IN_PROGRESS or body.status == ApiCallStatus.QUEUED
    ) and time.time() - start_time < 120:
        result_status = client.ml.get_text_to_cad_model_for_user(
            id=body.id,
        )

        body = result_status.root  # type: ignore

    assert body.status == ApiCallStatus.COMPLETED



from typing import Optional

from .base import KittyCadBaseModel


class Error(KittyCadBaseModel):
    """Error information from a response."""

    error_code: Optional[str] = None

    message: str

    request_id: str



from ..models.user_org_role import UserOrgRole
from .base import KittyCadBaseModel


class AddOrgMember(KittyCadBaseModel):
    """Data for adding a member to an org."""

    email: str

    role: UserOrgRole



from ..models.axis_direction_pair import AxisDirectionPair
from .base import KittyCadBaseModel


class System(KittyCadBaseModel):
    """Co-ordinate system definition.

    The `up` axis must be orthogonal to the `forward` axis.

    See [cglearn.eu] for background reading.

    [cglearn.eu](https://cglearn.eu/pub/computer-graphics/introduction-to-geometry#material-coordinate-systems-1)"""

    forward: AxisDirectionPair

    up: AxisDirectionPair



"""File input normalization helpers for the KittyCAD SDK.

This module provides utilities to normalize various file input types into
consistent file-like objects that can be used with HTTPX multipart uploads
and raw binary uploads.
"""

from __future__ import annotations

import io
import mimetypes
from pathlib import Path
from typing import IO, Optional, Tuple, cast

from ._io_types import SyncUpload


def _open_if_path(
    upload: SyncUpload,
) -> Tuple[IO[bytes], bool, Optional[str]]:
    """Normalize any upload input to (file_obj, should_close, filename).

    Args:
        upload: Any acceptable upload input

    Returns:
        A tuple of (file_object, should_close, filename)
        - file_object: A file-like object ready for reading
        - should_close: Whether we opened the file and should close it
        - filename: The filename if available (for multipart uploads)

    Note:
        Only close files that we opened. Don't close user-provided file objects.
    """
    # Handle path-like objects
    if isinstance(upload, (str, Path)) or hasattr(upload, "__fspath__"):
        path = Path(str(upload))  # Convert to str first for Path constructor
        filename = path.name
        file_obj = open(path, "rb")
        return file_obj, True, filename

    # Handle bytes-like objects
    if isinstance(upload, (bytes, bytearray, memoryview)):
        bytes_obj: IO[bytes] = io.BytesIO(bytes(upload))
        return bytes_obj, False, None

    # Handle iterables (streaming)
    if hasattr(upload, "__iter__") and not hasattr(upload, "read"):
        # Convert iterator to BytesIO for now
        # In a full implementation, this would be optimized for streaming
        chunks = []
        for chunk in upload:
            if isinstance(chunk, bytes):
                chunks.append(chunk)
            else:
                raise TypeError(f"Iterator must yield bytes, got {type(chunk)}")

        data = b"".join(chunks)
        iter_obj: IO[bytes] = io.BytesIO(data)
        return iter_obj, False, None

    # Handle file-like objects (already open)
    if hasattr(upload, "read"):
        filename_attr = getattr(upload, "name", None)
        if filename_attr and not isinstance(filename_attr, str):
            filename_attr = None  # Don't use non-string names
        return cast(IO[bytes], upload), False, filename_attr

    raise TypeError(f"Unsupported upload type: {type(upload)}")


def _get_content_type(
    filename: Optional[str] = None,
    content: Optional[bytes] = None,
    override: Optional[str] = None,
) -> str:
    """Guess content type from filename or content.

    Args:
        filename: Optional filename to guess from
        content: Optional content bytes to analyze
        override: Explicit content type override

    Returns:
        Content type string
    """
    if override:
        return override

    # Try to guess from filename first
    if filename:
        content_type, _ = mimetypes.guess_type(filename)
        if content_type:
            return content_type

    # Try to detect from content for common formats
    if content:
        # Check for common binary formats by magic bytes
        if content.startswith(b"\x89PNG\r\n\x1a\n"):
            return "image/png"
        elif content.startswith(b"\xff\xd8\xff"):
            return "image/jpeg"
        elif content.startswith(b"GIF8"):
            return "image/gif"
        elif content.startswith(b"\x00\x00\x00\x20ftyp"):
            return "video/mp4"
        elif content.startswith(b"%PDF"):
            return "application/pdf"
        elif content.startswith(b"PK\x03\x04"):
            return "application/zip"

    # Default to binary
    return "application/octet-stream"


def _should_use_multipart(
    upload: SyncUpload,
    filename: Optional[str] = None,
    content_type: Optional[str] = None,
    force_multipart: bool = False,
) -> bool:
    """Determine if upload should use multipart/form-data vs raw binary.

    Args:
        upload: The upload input
        filename: Optional filename
        content_type: Optional content type
        force_multipart: Force multipart even without filename

    Returns:
        True if should use multipart, False for raw binary

    Note:
        - Use multipart when we have a filename (file upload semantics)
        - Use multipart when explicitly forced
        - Use raw binary for bytes/streams without filename (data upload semantics)
    """
    if force_multipart:
        return True

    # If we have a filename, prefer multipart (file upload semantics)
    if filename:
        return True

    # For path-like objects, always use multipart
    if isinstance(upload, (str, Path)) or hasattr(upload, "__fspath__"):
        return True

    # For bytes-like without filename, use raw binary (data semantics)
    if isinstance(upload, (bytes, bytearray, memoryview)):
        return False

    # For file objects, check if they have a name
    if hasattr(upload, "read"):
        obj_filename = getattr(upload, "name", None)
        if obj_filename and isinstance(obj_filename, str):
            return True

    # Default to raw binary for iterators and unnamed objects
    return False


def _peek_file_content(file_obj: IO[bytes], peek_size: int = 512) -> bytes:
    """Peek at the beginning of a file without consuming it.

    Args:
        file_obj: File-like object to peek at
        peek_size: Number of bytes to peek

    Returns:
        The peeked bytes

    Note:
        This function ensures the file position is restored after peeking.
    """
    if not hasattr(file_obj, "seek") or not hasattr(file_obj, "tell"):
        # Can't seek, just return empty
        return b""

    current_pos = file_obj.tell()
    try:
        content = file_obj.read(peek_size)
        file_obj.seek(current_pos)
        return content
    except (OSError, io.UnsupportedOperation):
        # Reset position and return empty if seeking fails
        try:
            file_obj.seek(current_pos)
        except (OSError, io.UnsupportedOperation):
            pass
        return b""


def prepare_upload_input(
    upload: SyncUpload,
    filename: Optional[str] = None,
    content_type: Optional[str] = None,
    force_multipart: bool = False,
) -> Tuple[IO[bytes], bool, Optional[str], str, bool]:
    """Prepare any upload input for HTTP requests.

    Args:
        upload: Any acceptable upload input type
        filename: Override filename (for multipart)
        content_type: Override content type
        force_multipart: Force multipart even without filename

    Returns:
        Tuple of (file_obj, should_close, filename, content_type, use_multipart)

    Example:
        >>> from io import BytesIO
        >>> file_obj, should_close, filename, content_type, use_multipart = prepare_upload_input(
        ...     BytesIO(b"image data")
        ... )
        >>> # Use file_obj for upload, close if should_close is True
    """
    # Normalize the input
    file_obj, should_close, detected_filename = _open_if_path(upload)

    # Use provided filename or detected filename
    final_filename = filename or detected_filename

    # Peek at content for type detection if needed
    content_sample = (
        _peek_file_content(file_obj)
        if not content_type and hasattr(file_obj, "read")
        else b""
    )

    # Determine content type
    final_content_type = _get_content_type(
        filename=final_filename, content=content_sample, override=content_type
    )

    # Decide on upload method
    use_multipart = _should_use_multipart(
        upload=upload,
        filename=final_filename,
        content_type=final_content_type,
        force_multipart=force_multipart,
    )

    return file_obj, should_close, final_filename, final_content_type, use_multipart



from typing import Optional

from .base import KittyCadBaseModel


class HighlightSetEntity(KittyCadBaseModel):
    """The response from the `HighlightSetEntity` command."""

    entity_id: Optional[str] = None

    sequence: Optional[int] = None



from typing import Dict, Literal, Optional, Union

from pydantic import Field, RootModel
from typing_extensions import Annotated

from .base import KittyCadBaseModel


class OptionTextToCad(KittyCadBaseModel):
    """Response from the `TextToCad` tool."""

    error: Optional[str] = None

    outputs: Optional[Dict[str, str]] = None

    project_name: Optional[str] = None

    status_code: int

    type: Literal["text_to_cad"] = "text_to_cad"


class OptionEditKclCode(KittyCadBaseModel):
    """Response from the `EditKclCode` tool."""

    error: Optional[str] = None

    outputs: Optional[Dict[str, str]] = None

    project_name: Optional[str] = None

    status_code: int

    type: Literal["edit_kcl_code"] = "edit_kcl_code"


class OptionMechanicalKnowledgeBase(KittyCadBaseModel):
    """Mechanical knowledge base response."""

    response: str

    type: Literal["mechanical_knowledge_base"] = "mechanical_knowledge_base"


MlToolResult = RootModel[
    Annotated[
        Union[
            OptionTextToCad,
            OptionEditKclCode,
            OptionMechanicalKnowledgeBase,
        ],
        Field(discriminator="type"),
    ]
]



from ..models.oauth2_grant_type import OAuth2GrantType
from .base import KittyCadBaseModel


class DeviceAccessTokenRequestForm(KittyCadBaseModel):
    """The form for a device access token request."""

    client_id: str

    device_code: str

    grant_type: OAuth2GrantType



from enum import Enum


class CameraMovement(str, Enum):
    """A type of camera movement applied after certain camera operations"""  # noqa: E501

    """# Adjusts the camera position during the camera operation"""  # noqa: E501

    VANTAGE = "vantage"

    """# Keeps the camera position in place"""  # noqa: E501

    NONE = "none"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class ModelingAppSubscriptionTierName(str, Enum):
    """An enum representing a Modeling App subscription tier name."""  # noqa: E501

    """# The free tier."""  # noqa: E501

    FREE = "free"

    """# The plus tier."""  # noqa: E501

    PLUS = "plus"

    """# The pro tier."""  # noqa: E501

    PRO = "pro"

    """# The team tier."""  # noqa: E501

    TEAM = "team"

    """# The enterprise tier."""  # noqa: E501

    ENTERPRISE = "enterprise"

    def __str__(self) -> str:
        return str(self.value)



import datetime
from typing import Optional

from ..models.api_call_status import ApiCallStatus
from ..models.file_import_format import FileImportFormat
from ..models.unit_density import UnitDensity
from ..models.unit_mass import UnitMass
from ..models.uuid import Uuid
from .base import KittyCadBaseModel


class FileMass(KittyCadBaseModel):
    """A file mass result."""

    completed_at: Optional[datetime.datetime] = None

    created_at: datetime.datetime

    error: Optional[str] = None

    id: Uuid

    mass: Optional[float] = None

    material_density: float = 0.0

    material_density_unit: UnitDensity

    output_unit: UnitMass

    src_format: FileImportFormat

    started_at: Optional[datetime.datetime] = None

    status: ApiCallStatus

    updated_at: datetime.datetime

    user_id: Uuid



from enum import Enum


class CodeLanguage(str, Enum):
    """The language code is written in.

    <details><summary>JSON schema</summary>

    ```json { "description": "The language code is written in.", "oneOf": [ { "description": "The `go` programming language.", "type": "string", "enum": [ "go" ] }, { "description": "The `python` programming language.", "type": "string", "enum": [ "python" ] }, { "description": "The `node` programming language.", "type": "string", "enum": [ "node" ] } ] } ``` </details>"""  # noqa: E501

    """# The `go` programming language."""  # noqa: E501

    GO = "go"

    """# The `python` programming language."""  # noqa: E501

    PYTHON = "python"

    """# The `node` programming language."""  # noqa: E501

    NODE = "node"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class RtcSdpType(str, Enum):
    """SDPType describes the type of an SessionDescription."""  # noqa: E501

    """# Unspecified indicates that the type is unspecified."""  # noqa: E501

    UNSPECIFIED = "unspecified"

    """# indicates that a description MUST be treated as an SDP offer."""  # noqa: E501

    OFFER = "offer"

    """# indicates that a description MUST be treated as an SDP answer, but not a final answer. A description used as an SDP pranswer may be applied as a response to an SDP offer, or an update to a previously sent SDP pranswer."""  # noqa: E501

    PRANSWER = "pranswer"

    """# indicates that a description MUST be treated as an SDP final answer, and the offer-answer exchange MUST be considered complete. A description used as an SDP answer may be applied as a response to an SDP offer or as an update to a previously sent SDP pranswer."""  # noqa: E501

    ANSWER = "answer"

    """# indicates that a description MUST be treated as canceling the current SDP negotiation and moving the SDP offer and answer back to what it was in the previous stable state. Note the local or remote SDP descriptions in the previous stable state could be null if there has not yet been a successful offer-answer negotiation."""  # noqa: E501

    ROLLBACK = "rollback"

    def __str__(self) -> str:
        return str(self.value)



from typing import Optional

from ..models.inquiry_type import InquiryType
from .base import KittyCadBaseModel


class InquiryForm(KittyCadBaseModel):
    """The form for a public inquiry submission."""

    company: Optional[str] = None

    email: str

    first_name: str

    industry: Optional[str] = None

    inquiry_type: InquiryType

    last_name: str

    message: str

    phone: Optional[str] = None



from enum import Enum


class UnitForce(str, Enum):
    """The valid types of force units."""  # noqa: E501

    """# Dynes <https://en.wikipedia.org/wiki/Dyne>"""  # noqa: E501

    DYNES = "dynes"

    """# Kiloponds <https://en.wikipedia.org/wiki/Kilopond>"""  # noqa: E501

    KILOPONDS = "kiloponds"

    """# Micronewtons <https://en.wikipedia.org/wiki/Newton_(unit)>"""  # noqa: E501

    MICRONEWTONS = "micronewtons"

    """# Millinewtons <https://en.wikipedia.org/wiki/Newton_(unit)>"""  # noqa: E501

    MILLINEWTONS = "millinewtons"

    """# Newtons <https://en.wikipedia.org/wiki/Newton_(unit)>"""  # noqa: E501

    NEWTONS = "newtons"

    """# Poundals <https://en.wikipedia.org/wiki/Poundal>"""  # noqa: E501

    POUNDALS = "poundals"

    """# Pounds <https://en.wikipedia.org/wiki/Pound_(force)>"""  # noqa: E501

    POUNDS = "pounds"

    def __str__(self) -> str:
        return str(self.value)



from ..models.angle import Angle
from ..models.origin_type import OriginType
from ..models.point3d import Point3d
from .base import KittyCadBaseModel


class Rotation(KittyCadBaseModel):
    """A rotation defined by an axis, origin of rotation, and an angle."""

    angle: Angle

    axis: Point3d

    origin: OriginType



from enum import Enum


class StlStorage(str, Enum):
    """Export storage."""  # noqa: E501

    """# Plaintext encoding."""  # noqa: E501

    ASCII = "ascii"

    """# Binary STL encoding.

This is the default setting."""  # noqa: E501

    BINARY = "binary"

    def __str__(self) -> str:
        return str(self.value)



from typing import Optional

from .base import KittyCadBaseModel


class AuthCallback(KittyCadBaseModel):
    """The authentication callback from the OAuth 2.0 client. This is typically posted to the redirect URL as query params after authenticating."""

    code: Optional[str] = None

    id_token: Optional[str] = None

    state: Optional[str] = None

    user: Optional[str] = None



"""Raw binary upload helpers for the KittyCAD SDK.

This module provides utilities for uploading raw binary data (application/octet-stream)
without multipart encoding, which is useful for APIs that expect raw file content.
"""

from __future__ import annotations

from typing import IO, Any, Dict, Optional, cast

import httpx

from ._file_inputs import prepare_upload_input
from ._io_types import ProgressCallback, SyncUpload
from ._progress import wrap_with_progress


def create_binary_upload(
    file_input: SyncUpload,
    content_type: Optional[str] = None,
    progress_callback: Optional[ProgressCallback] = None,
) -> tuple[bytes, str]:
    """Create a binary upload from file input.

    Args:
        file_input: The file content (path, file object, bytes, etc.)
        content_type: Override content type (defaults to application/octet-stream)
        progress_callback: Optional progress tracking callback

    Returns:
        Tuple of (binary_data, content_type)

    Example:
        >>> from io import BytesIO
        >>> data, content_type = create_binary_upload(BytesIO(b"image data"))
        >>> # response = httpx.post(url, content=data, headers={"Content-Type": content_type})
    """
    # Prepare the file input (force non-multipart)
    file_obj, should_close, _, detected_content_type, _ = prepare_upload_input(
        upload=file_input, content_type=content_type, force_multipart=False
    )

    try:
        # Wrap with progress tracking if requested
        if progress_callback and hasattr(file_obj, "read"):
            wrapped_file_obj = wrap_with_progress(
                file_obj=file_obj, progress_callback=progress_callback, for_reading=True
            )
            if hasattr(wrapped_file_obj, "read"):
                file_obj = cast(IO[bytes], wrapped_file_obj)

        # Read all data
        if hasattr(file_obj, "read"):
            binary_data = file_obj.read()
        else:
            raise TypeError(f"Object does not support reading: {type(file_obj)}")

        # Use provided content type or detected type, fallback to octet-stream
        final_content_type = (
            content_type or detected_content_type or "application/octet-stream"
        )

        return binary_data, final_content_type

    finally:
        # Clean up if we opened the file
        if should_close and hasattr(file_obj, "close"):
            try:
                file_obj.close()
            except Exception:
                pass


def create_binary_stream(
    file_input: SyncUpload,
    content_type: Optional[str] = None,
    progress_callback: Optional[ProgressCallback] = None,
    chunk_size: int = 8192,
):
    """Create a streaming binary upload from file input.

    Args:
        file_input: The file content (path, file object, bytes, etc.)
        content_type: Override content type (defaults to application/octet-stream)
        progress_callback: Optional progress tracking callback
        chunk_size: Size of chunks to read when streaming

    Returns:
        Tuple of (stream_generator, content_type, cleanup_func)

    Example:
        >>> from io import BytesIO
        >>> stream, content_type, cleanup = create_binary_stream(BytesIO(b"large file data"))
        >>> try:
        ...     # Use stream and content_type for HTTP request
        ...     pass
        ... finally:
        ...     cleanup()
    """
    # Handle iterators specially for streaming (but not strings/paths)
    if (
        hasattr(file_input, "__iter__")
        and not hasattr(file_input, "read")
        and not isinstance(file_input, (str, bytes, bytearray, memoryview))
        and not hasattr(file_input, "__fspath__")
    ):
        # For iterators, pass them through directly for streaming
        def iterator_stream_generator():
            """Generator that yields chunks from the iterator."""
            for chunk in file_input:
                if isinstance(chunk, bytes):
                    yield chunk
                else:
                    raise TypeError(f"Iterator must yield bytes, got {type(chunk)}")

        def iterator_cleanup_func():
            # Nothing to clean up for iterators
            pass

        final_content_type = content_type or "application/octet-stream"
        return iterator_stream_generator(), final_content_type, iterator_cleanup_func

    # Prepare the file input for non-iterators
    file_obj, should_close, _, detected_content_type, _ = prepare_upload_input(
        upload=file_input, content_type=content_type, force_multipart=False
    )

    # Wrap with progress tracking if requested
    if progress_callback and hasattr(file_obj, "read"):
        wrapped_file_obj = wrap_with_progress(
            file_obj=file_obj, progress_callback=progress_callback, for_reading=True
        )
        if hasattr(wrapped_file_obj, "read"):
            file_obj = cast(IO[bytes], wrapped_file_obj)

    def stream_generator():
        """Generator that yields chunks of the file."""
        try:
            while True:
                if hasattr(file_obj, "read"):
                    chunk = file_obj.read(chunk_size)
                    if not chunk:
                        break
                    yield chunk
                else:
                    break
        except Exception:
            # If streaming fails, clean up and re-raise
            cleanup_func()
            raise

    def cleanup_func():
        """Clean up resources."""
        if should_close and hasattr(file_obj, "close"):
            try:
                file_obj.close()
            except Exception:
                pass

    # Use provided content type or detected type, fallback to octet-stream
    final_content_type = (
        content_type or detected_content_type or "application/octet-stream"
    )

    return stream_generator(), final_content_type, cleanup_func


class BinaryUploadContext:
    """Context manager for binary uploads with automatic cleanup.

    Example:
        >>> from io import BytesIO
        >>> with BinaryUploadContext(BytesIO(b"test data")) as upload:
        ...     # Use upload.data and upload.headers for HTTP request
        ...     pass
    """

    def __init__(
        self,
        file_input: SyncUpload,
        content_type: Optional[str] = None,
        progress_callback: Optional[ProgressCallback] = None,
        stream: bool = False,
        chunk_size: int = 8192,
    ):
        """Initialize the binary upload context.

        Args:
            file_input: The file content (path, file object, bytes, etc.)
            content_type: Override content type
            progress_callback: Optional progress tracking callback
            stream: Whether to use streaming (True) or load all data (False)
            chunk_size: Size of chunks when streaming
        """
        self.file_input = file_input
        self.content_type = content_type
        self.progress_callback = progress_callback
        self.stream = stream
        self.chunk_size = chunk_size
        self.data: Optional[Any] = None
        self.headers: Dict[str, str] = {}
        self._cleanup_func = None

    def __enter__(self):
        """Prepare the binary upload."""
        if self.stream:
            # Use streaming approach
            stream_gen, content_type, cleanup_func = create_binary_stream(
                file_input=self.file_input,
                content_type=self.content_type,
                progress_callback=self.progress_callback,
                chunk_size=self.chunk_size,
            )
            self.data = stream_gen
            self._cleanup_func = cleanup_func
        else:
            # Load all data at once
            data, content_type = create_binary_upload(
                file_input=self.file_input,
                content_type=self.content_type,
                progress_callback=self.progress_callback,
            )
            self.data = data

        self.headers = {"Content-Type": content_type}
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Clean up resources."""
        if self._cleanup_func:
            self._cleanup_func()
        return None


def upload_file_binary(
    client: httpx.Client,
    url: str,
    file_input: SyncUpload,
    method: str = "POST",
    content_type: Optional[str] = None,
    progress_callback: Optional[ProgressCallback] = None,
    stream: bool = False,
    chunk_size: int = 8192,
    **request_kwargs,
) -> httpx.Response:
    """Upload a file as raw binary data with automatic cleanup.

    Args:
        client: HTTPX client to use for the request
        url: URL to upload to
        file_input: The file content (path, file object, bytes, etc.)
        method: HTTP method (default: POST)
        content_type: Override content type
        progress_callback: Optional progress tracking callback
        stream: Whether to use streaming upload
        chunk_size: Size of chunks when streaming
        **request_kwargs: Additional arguments for the HTTP request

    Returns:
        The HTTP response

    Example:
        >>> from io import BytesIO
        >>> import httpx
        >>> response = upload_file_binary(
        ...     client=httpx.Client(),
        ...     url="https://example.com/upload-binary",
        ...     file_input=BytesIO(b"file data"),
        ... )
    """
    with BinaryUploadContext(
        file_input=file_input,
        content_type=content_type,
        progress_callback=progress_callback,
        stream=stream,
        chunk_size=chunk_size,
    ) as upload:
        # Make the request with binary content
        response = client.request(
            method=method,
            url=url,
            content=upload.data,
            headers={**upload.headers, **request_kwargs.pop("headers", {})},
            **request_kwargs,
        )
        return response


async def upload_file_binary_async(
    client: httpx.AsyncClient,
    url: str,
    file_input: SyncUpload,
    method: str = "POST",
    content_type: Optional[str] = None,
    progress_callback: Optional[ProgressCallback] = None,
    stream: bool = False,
    chunk_size: int = 8192,
    **request_kwargs,
) -> httpx.Response:
    """Async version of upload_file_binary.

    Args:
        client: Async HTTPX client to use for the request
        url: URL to upload to
        file_input: The file content (path, file object, bytes, etc.)
        method: HTTP method (default: POST)
        content_type: Override content type
        progress_callback: Optional progress tracking callback
        stream: Whether to use streaming upload
        chunk_size: Size of chunks when streaming
        **request_kwargs: Additional arguments for the HTTP request

    Returns:
        The HTTP response
    """
    with BinaryUploadContext(
        file_input=file_input,
        content_type=content_type,
        progress_callback=progress_callback,
        stream=stream,
        chunk_size=chunk_size,
    ) as upload:
        # Make the async request with binary content
        response = await client.request(
            method=method,
            url=url,
            content=upload.data,
            headers={**upload.headers, **request_kwargs.pop("headers", {})},
            **request_kwargs,
        )
        return response


def prepare_binary_headers(
    content_type: Optional[str] = None,
    content_length: Optional[int] = None,
    additional_headers: Optional[Dict[str, str]] = None,
) -> Dict[str, str]:
    """Prepare headers for binary uploads.

    Args:
        content_type: Content type for the upload
        content_length: Content length if known
        additional_headers: Additional headers to include

    Returns:
        Dictionary of headers for the request
    """
    headers = {}

    if content_type:
        headers["Content-Type"] = content_type
    else:
        headers["Content-Type"] = "application/octet-stream"

    if content_length is not None:
        headers["Content-Length"] = str(content_length)

    if additional_headers:
        headers.update(additional_headers)

    return headers



"""Tests for pagination functionality."""

from typing import AsyncIterator, Iterator, List, Optional
from unittest.mock import AsyncMock, Mock

import pytest
from pydantic import BaseModel

from kittycad import AsyncKittyCAD, KittyCAD
from kittycad.pagination import AsyncPageIterator, SyncPageIterator


# Mock data models for testing
class MockItem(BaseModel):
    id: str
    name: str


class MockPage(BaseModel):
    items: List[MockItem]
    next_page: Optional[str] = None


def test_sync_page_iterator_single_page():
    """Test sync pagination with a single page."""
    # Create mock data
    items = [
        MockItem(id="1", name="Item 1"),
        MockItem(id="2", name="Item 2"),
    ]
    page = MockPage(items=items, next_page=None)

    # Create mock page fetcher
    mock_fetcher = Mock(return_value=page)

    # Create iterator
    iterator = SyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"limit": 10},
        item_type=MockItem,
    )

    # Collect all items
    result_items: list[MockItem] = list(iterator)

    # Verify results
    assert len(result_items) == 2
    assert result_items[0].id == "1"
    assert result_items[0].name == "Item 1"
    assert result_items[1].id == "2"
    assert result_items[1].name == "Item 2"

    # Verify fetcher was called once with initial kwargs
    mock_fetcher.assert_called_once_with(limit=10)


def test_sync_page_iterator_multiple_pages():
    """Test sync pagination with multiple pages."""
    # Create mock data for multiple pages
    page1_items = [MockItem(id="1", name="Item 1"), MockItem(id="2", name="Item 2")]
    page2_items = [MockItem(id="3", name="Item 3"), MockItem(id="4", name="Item 4")]
    page3_items = [MockItem(id="5", name="Item 5")]

    page1 = MockPage(items=page1_items, next_page="token2")
    page2 = MockPage(items=page2_items, next_page="token3")
    page3 = MockPage(items=page3_items, next_page=None)

    # Create mock page fetcher that returns different pages based on token
    def mock_fetch_page(**kwargs):
        if "page_token" not in kwargs:
            return page1
        elif kwargs["page_token"] == "token2":
            return page2
        elif kwargs["page_token"] == "token3":
            return page3
        else:
            raise ValueError(f"Unexpected page token: {kwargs['page_token']}")

    mock_fetcher = Mock(side_effect=mock_fetch_page)

    # Create iterator
    iterator = SyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"limit": 2},
        item_type=MockItem,
    )

    # Collect all items
    result_items: list[MockItem] = list(iterator)

    # Verify results
    assert len(result_items) == 5
    assert [item.id for item in result_items] == ["1", "2", "3", "4", "5"]
    assert [item.name for item in result_items] == [
        "Item 1",
        "Item 2",
        "Item 3",
        "Item 4",
        "Item 5",
    ]

    # Verify fetcher was called three times with correct arguments
    assert mock_fetcher.call_count == 3
    mock_fetcher.assert_any_call(limit=2)  # First call
    mock_fetcher.assert_any_call(limit=2, page_token="token2")  # Second call
    mock_fetcher.assert_any_call(limit=2, page_token="token3")  # Third call


def test_sync_page_iterator_empty_page():
    """Test sync pagination with empty pages."""
    empty_page = MockPage(items=[], next_page=None)
    mock_fetcher = Mock(return_value=empty_page)

    iterator = SyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={},
        item_type=MockItem,
    )

    result_items: list[MockItem] = list(iterator)

    assert len(result_items) == 0
    mock_fetcher.assert_called_once_with()


@pytest.mark.asyncio
async def test_async_page_iterator_single_page():
    """Test async pagination with a single page."""
    # Create mock data
    items = [
        MockItem(id="1", name="Item 1"),
        MockItem(id="2", name="Item 2"),
    ]
    page = MockPage(items=items, next_page=None)

    # Create async mock page fetcher
    mock_fetcher = AsyncMock(return_value=page)

    # Create async iterator
    iterator = AsyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"limit": 10},
        item_type=MockItem,
    )

    # Collect all items
    result_items: list[MockItem] = []
    item: MockItem
    async for item in iterator:
        result_items.append(item)

    # Verify results
    assert len(result_items) == 2
    assert result_items[0].id == "1"
    assert result_items[0].name == "Item 1"
    assert result_items[1].id == "2"
    assert result_items[1].name == "Item 2"

    # Verify fetcher was called once with initial kwargs
    mock_fetcher.assert_called_once_with(limit=10)


@pytest.mark.asyncio
async def test_async_page_iterator_multiple_pages():
    """Test async pagination with multiple pages."""
    # Create mock data for multiple pages
    page1_items = [MockItem(id="1", name="Item 1"), MockItem(id="2", name="Item 2")]
    page2_items = [MockItem(id="3", name="Item 3"), MockItem(id="4", name="Item 4")]
    page3_items = [MockItem(id="5", name="Item 5")]

    page1 = MockPage(items=page1_items, next_page="token2")
    page2 = MockPage(items=page2_items, next_page="token3")
    page3 = MockPage(items=page3_items, next_page=None)

    # Create async mock page fetcher that returns different pages based on token
    async def mock_fetch_page(**kwargs):
        if "page_token" not in kwargs:
            return page1
        elif kwargs["page_token"] == "token2":
            return page2
        elif kwargs["page_token"] == "token3":
            return page3
        else:
            raise ValueError(f"Unexpected page token: {kwargs['page_token']}")

    mock_fetcher = AsyncMock(side_effect=mock_fetch_page)

    # Create async iterator
    iterator = AsyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"limit": 2},
        item_type=MockItem,
    )

    # Collect all items
    result_items: list[MockItem] = []
    item: MockItem
    async for item in iterator:
        result_items.append(item)

    # Verify results
    assert len(result_items) == 5
    assert [item.id for item in result_items] == ["1", "2", "3", "4", "5"]
    assert [item.name for item in result_items] == [
        "Item 1",
        "Item 2",
        "Item 3",
        "Item 4",
        "Item 5",
    ]

    # Verify fetcher was called three times with correct arguments
    assert mock_fetcher.call_count == 3
    mock_fetcher.assert_any_call(limit=2)  # First call
    mock_fetcher.assert_any_call(limit=2, page_token="token2")  # Second call
    mock_fetcher.assert_any_call(limit=2, page_token="token3")  # Third call


@pytest.mark.asyncio
async def test_async_page_iterator_empty_page():
    """Test async pagination with empty pages."""
    empty_page = MockPage(items=[], next_page=None)
    mock_fetcher = AsyncMock(return_value=empty_page)

    iterator = AsyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={},
        item_type=MockItem,
    )

    result_items: list[MockItem] = []
    item: MockItem
    async for item in iterator:
        result_items.append(item)

    assert len(result_items) == 0
    mock_fetcher.assert_called_once_with()


def test_sync_page_iterator_preserves_initial_kwargs():
    """Test that initial kwargs are preserved across page requests."""
    page1 = MockPage(items=[MockItem(id="1", name="Item 1")], next_page="token2")
    page2 = MockPage(items=[MockItem(id="2", name="Item 2")], next_page=None)

    def mock_fetch_page(**kwargs):
        # Verify that sort_by is always passed
        assert kwargs["sort_by"] == "created_at"
        if "page_token" not in kwargs:
            return page1
        else:
            return page2

    mock_fetcher = Mock(side_effect=mock_fetch_page)

    iterator = SyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"sort_by": "created_at", "limit": 1},
        item_type=MockItem,
    )

    # Trigger pagination
    result_items: list[MockItem] = list(iterator)

    assert len(result_items) == 2
    assert mock_fetcher.call_count == 2


@pytest.mark.asyncio
async def test_async_page_iterator_preserves_initial_kwargs():
    """Test that async initial kwargs are preserved across page requests."""
    page1 = MockPage(items=[MockItem(id="1", name="Item 1")], next_page="token2")
    page2 = MockPage(items=[MockItem(id="2", name="Item 2")], next_page=None)

    async def mock_fetch_page(**kwargs):
        # Verify that sort_by is always passed
        assert kwargs["sort_by"] == "created_at"
        if "page_token" not in kwargs:
            return page1
        else:
            return page2

    mock_fetcher = AsyncMock(side_effect=mock_fetch_page)

    iterator = AsyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"sort_by": "created_at", "limit": 1},
        item_type=MockItem,
    )

    # Trigger pagination
    result_items: list[MockItem] = []
    item: MockItem
    async for item in iterator:
        result_items.append(item)

    assert len(result_items) == 2
    assert mock_fetcher.call_count == 2


def test_sync_page_iterator_handles_none_items():
    """Test sync pagination handles None items gracefully."""

    # Page with None items attribute
    class BadPage:
        items = None
        next_page = None

    mock_fetcher = Mock(return_value=BadPage())

    iterator = SyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={},
        item_type=MockItem,
    )

    result_items: list[MockItem] = list(iterator)
    assert len(result_items) == 0


@pytest.mark.asyncio
async def test_async_page_iterator_handles_none_items():
    """Test async pagination handles None items gracefully."""

    # Page with None items attribute
    class BadPage:
        items = None
        next_page = None

    mock_fetcher = AsyncMock(return_value=BadPage())

    iterator = AsyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={},
        item_type=MockItem,
    )

    result_items: list[MockItem] = []
    item: MockItem
    async for item in iterator:
        result_items.append(item)

    assert len(result_items) == 0


def test_sync_page_iterator_multiple_iteration():
    """Test that sync iterator can be iterated multiple times."""
    items = [MockItem(id="1", name="Item 1")]
    page = MockPage(items=items, next_page=None)
    mock_fetcher = Mock(return_value=page)

    iterator = SyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={},
        item_type=MockItem,
    )

    # First iteration
    result1: list[MockItem] = list(iterator)
    # Second iteration should start fresh
    result2: list[MockItem] = list(iterator)

    assert len(result1) == 1
    assert len(result2) == 1
    # Should have called fetcher twice (once per iteration)
    assert mock_fetcher.call_count == 2


@pytest.mark.asyncio
async def test_async_page_iterator_multiple_iteration():
    """Test that async iterator can be iterated multiple times."""
    items = [MockItem(id="1", name="Item 1")]
    page = MockPage(items=items, next_page=None)
    mock_fetcher = AsyncMock(return_value=page)

    iterator = AsyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={},
        item_type=MockItem,
    )

    # First iteration
    result1: list[MockItem] = []
    item: MockItem
    async for item in iterator:
        result1.append(item)

    # Second iteration should start fresh
    result2: list[MockItem] = []
    # Use same item variable for second iteration
    async for item in iterator:
        result2.append(item)

    assert len(result1) == 1
    assert len(result2) == 1
    # Should have called fetcher twice (once per iteration)
    assert mock_fetcher.call_count == 2


def test_sync_page_iterator_scan_immutability():
    """Test that scan parameters cannot be changed mid-iteration (scan immutability)."""
    page1 = MockPage(items=[MockItem(id="1", name="Item A")], next_page="token2")
    page2 = MockPage(items=[MockItem(id="2", name="Item B")], next_page=None)

    def mock_fetch_page(**kwargs):
        # Verify that sort_by remains consistent across pages
        assert kwargs["sort_by"] == "name_ascending", (
            f"Sort changed to {kwargs.get('sort_by')}"
        )
        if "page_token" not in kwargs:
            return page1
        else:
            return page2

    mock_fetcher = Mock(side_effect=mock_fetch_page)

    iterator = SyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"sort_by": "name_ascending", "limit": 1},
        item_type=MockItem,
    )

    # Start iteration
    iter_obj: Iterator[MockItem] = iter(iterator)
    first_item = next(iter_obj)
    assert first_item.id == "1"

    # Attempt to modify the iterator's initial kwargs (should not affect ongoing iteration)
    iterator._initial_kwargs["sort_by"] = "name_descending"

    # Continue iteration - should still use original sort
    second_item = next(iter_obj)
    assert second_item.id == "2"

    # Verify fetcher was called with original sort both times
    assert mock_fetcher.call_count == 2


@pytest.mark.asyncio
async def test_async_page_iterator_scan_immutability():
    """Test that async scan parameters cannot be changed mid-iteration."""
    page1 = MockPage(items=[MockItem(id="1", name="Item A")], next_page="token2")
    page2 = MockPage(items=[MockItem(id="2", name="Item B")], next_page=None)

    async def mock_fetch_page(**kwargs):
        # Verify that sort_by remains consistent across pages
        assert kwargs["sort_by"] == "name_ascending", (
            f"Sort changed to {kwargs.get('sort_by')}"
        )
        if "page_token" not in kwargs:
            return page1
        else:
            return page2

    mock_fetcher = AsyncMock(side_effect=mock_fetch_page)

    iterator = AsyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"sort_by": "name_ascending", "limit": 1},
        item_type=MockItem,
    )

    # Start async iteration
    aiter_obj: AsyncIterator[MockItem] = iterator.__aiter__()
    first_item = await aiter_obj.__anext__()
    assert first_item.id == "1"

    # Attempt to modify the iterator's initial kwargs (should not affect ongoing iteration)
    iterator._initial_kwargs["sort_by"] = "name_descending"

    # Continue iteration - should still use original sort
    second_item = await aiter_obj.__anext__()
    assert second_item.id == "2"

    # Verify fetcher was called with original sort both times
    assert mock_fetcher.call_count == 2


def test_sync_page_iterator_extra_params_carry_across_pages():
    """Test that extra required query params are carried across all pages."""
    page1 = MockPage(items=[MockItem(id="1", name="Item 1")], next_page="token2")
    page2 = MockPage(items=[MockItem(id="2", name="Item 2")], next_page=None)

    def mock_fetch_page(**kwargs):
        # Verify all extra params are present in every request
        assert kwargs["user_id"] == "user123"
        assert kwargs["org_id"] == "org456"
        assert kwargs["include_deleted"]
        if "page_token" not in kwargs:
            return page1
        else:
            return page2

    mock_fetcher = Mock(side_effect=mock_fetch_page)

    iterator = SyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={
            "user_id": "user123",
            "org_id": "org456",
            "include_deleted": True,
            "limit": 1,
        },
        item_type=MockItem,
    )

    result_items: list[MockItem] = list(iterator)

    assert len(result_items) == 2
    assert mock_fetcher.call_count == 2


@pytest.mark.asyncio
async def test_async_page_iterator_extra_params_carry_across_pages():
    """Test that async extra required query params are carried across all pages."""
    page1 = MockPage(items=[MockItem(id="1", name="Item 1")], next_page="token2")
    page2 = MockPage(items=[MockItem(id="2", name="Item 2")], next_page=None)

    async def mock_fetch_page(**kwargs):
        # Verify all extra params are present in every request
        assert kwargs["user_id"] == "user123"
        assert kwargs["org_id"] == "org456"
        assert kwargs["include_deleted"]
        if "page_token" not in kwargs:
            return page1
        else:
            return page2

    mock_fetcher = AsyncMock(side_effect=mock_fetch_page)

    iterator = AsyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={
            "user_id": "user123",
            "org_id": "org456",
            "include_deleted": True,
            "limit": 1,
        },
        item_type=MockItem,
    )

    result_items: list[MockItem] = []
    item: MockItem
    async for item in iterator:
        result_items.append(item)

    assert len(result_items) == 2
    assert mock_fetcher.call_count == 2


def test_sync_page_iterator_memory_efficiency():
    """Test that sync pagination uses O(1) memory (yields as we fetch)."""
    # Create a large number of pages with small page sizes to test memory efficiency
    num_pages = 100
    items_per_page = 10

    page_call_count = 0

    def mock_fetch_page(**kwargs):
        nonlocal page_call_count
        page_call_count += 1

        # Generate items for this page
        start_id = (page_call_count - 1) * items_per_page + 1
        items = [
            MockItem(id=str(start_id + i), name=f"Item {start_id + i}")
            for i in range(items_per_page)
        ]

        # Determine next page token
        next_page = (
            f"token{page_call_count + 1}" if page_call_count < num_pages else None
        )

        return MockPage(items=items, next_page=next_page)

    mock_fetcher = Mock(side_effect=mock_fetch_page)

    iterator = SyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"limit": items_per_page},
        item_type=MockItem,
    )

    # Process items one by one and verify they come in order
    item_count = 0
    item: MockItem
    for item in iterator:
        item_count += 1
        # Mypy needs help with early break - cast to avoid Never type
        typed_item: MockItem = item
        assert typed_item.id == str(item_count)
        assert typed_item.name == f"Item {item_count}"

        # Stop after a reasonable number to avoid excessive test time
        if item_count >= 50:
            break

    # Verify that we only fetched the pages we needed (not all 100)
    expected_pages = (50 // items_per_page) + (1 if 50 % items_per_page > 0 else 0)
    assert page_call_count == expected_pages
    assert item_count == 50


@pytest.mark.asyncio
async def test_async_page_iterator_memory_efficiency():
    """Test that async pagination uses O(1) memory (yields as we fetch)."""
    # Create a large number of pages with small page sizes to test memory efficiency
    num_pages = 100
    items_per_page = 10

    page_call_count = 0

    async def mock_fetch_page(**kwargs):
        nonlocal page_call_count
        page_call_count += 1

        # Generate items for this page
        start_id = (page_call_count - 1) * items_per_page + 1
        items = [
            MockItem(id=str(start_id + i), name=f"Item {start_id + i}")
            for i in range(items_per_page)
        ]

        # Determine next page token
        next_page = (
            f"token{page_call_count + 1}" if page_call_count < num_pages else None
        )

        return MockPage(items=items, next_page=next_page)

    mock_fetcher = AsyncMock(side_effect=mock_fetch_page)

    iterator = AsyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"limit": items_per_page},
        item_type=MockItem,
    )

    # Process items one by one and verify they come in order
    item_count = 0
    item: MockItem
    async for item in iterator:
        item_count += 1
        # Mypy needs help with early break - cast to avoid Never type
        typed_item: MockItem = item
        assert typed_item.id == str(item_count)
        assert typed_item.name == f"Item {item_count}"

        # Stop after a reasonable number to avoid excessive test time
        if item_count >= 50:
            break

    # Verify that we only fetched the pages we needed (not all 100)
    expected_pages = (50 // items_per_page) + (1 if 50 % items_per_page > 0 else 0)
    assert page_call_count == expected_pages
    assert item_count == 50


def test_sync_page_iterator_large_dataset():
    """Test sync iteration over a large dataset to verify N items are yielded."""
    total_items = 257  # Non-round number to test edge cases
    items_per_page = 25

    def mock_fetch_page(**kwargs):
        # Determine which page we're on based on page_token
        if "page_token" not in kwargs:
            page_num = 0
        else:
            page_num = int(kwargs["page_token"].replace("page", "")) - 1

        # Calculate items for this page
        start_id = page_num * items_per_page + 1
        end_id = min(start_id + items_per_page, total_items + 1)

        items = [MockItem(id=str(i), name=f"Item {i}") for i in range(start_id, end_id)]

        # Determine next page
        next_page = f"page{page_num + 2}" if end_id <= total_items else None

        return MockPage(items=items, next_page=next_page)

    mock_fetcher = Mock(side_effect=mock_fetch_page)

    iterator = SyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"limit": items_per_page},
        item_type=MockItem,
    )

    # Collect all items
    result_items: list[MockItem] = list(iterator)

    # Verify we got exactly the expected number of items
    assert len(result_items) == total_items

    # Verify all items are in order
    for i, item in enumerate(result_items, 1):
        assert item.id == str(i)
        assert item.name == f"Item {i}"

    # Verify correct number of page fetches
    expected_pages = (total_items + items_per_page - 1) // items_per_page
    assert mock_fetcher.call_count == expected_pages


@pytest.mark.asyncio
async def test_async_page_iterator_large_dataset():
    """Test async iteration over a large dataset to verify N items are yielded."""
    total_items = 257  # Non-round number to test edge cases
    items_per_page = 25

    async def mock_fetch_page(**kwargs):
        # Determine which page we're on based on page_token
        if "page_token" not in kwargs:
            page_num = 0
        else:
            page_num = int(kwargs["page_token"].replace("page", "")) - 1

        # Calculate items for this page
        start_id = page_num * items_per_page + 1
        end_id = min(start_id + items_per_page, total_items + 1)

        items = [MockItem(id=str(i), name=f"Item {i}") for i in range(start_id, end_id)]

        # Determine next page
        next_page = f"page{page_num + 2}" if end_id <= total_items else None

        return MockPage(items=items, next_page=next_page)

    mock_fetcher = AsyncMock(side_effect=mock_fetch_page)

    iterator = AsyncPageIterator(
        page_fetcher=mock_fetcher,
        initial_kwargs={"limit": items_per_page},
        item_type=MockItem,
    )

    # Collect all items
    result_items: list[MockItem] = []
    item: MockItem
    async for item in iterator:
        result_items.append(item)

    # Verify we got exactly the expected number of items
    assert len(result_items) == total_items

    # Verify all items are in order
    for i, item in enumerate(result_items, 1):
        assert item.id == str(i)
        assert item.name == f"Item {i}"

    # Verify correct number of page fetches
    expected_pages = (total_items + items_per_page - 1) // items_per_page
    assert mock_fetcher.call_count == expected_pages


# Integration tests with real API calls
def test_sync_pagination_integration_text_to_cad():
    """Integration test for sync pagination with real text-to-cad API calls."""
    # Create client with real API token
    client = KittyCAD()

    # Call paginated endpoint - now returns SyncPageIterator
    iterator = client.ml.list_text_to_cad_models_for_user(  # type: ignore[attr-defined]
        limit=10,  # Small page size to test pagination
    )

    # Verify we got a SyncPageIterator
    from kittycad.models.text_to_cad_response import TextToCadResponse
    from kittycad.pagination import SyncPageIterator

    assert isinstance(iterator, SyncPageIterator), (
        f"Expected SyncPageIterator, got {type(iterator)}"
    )

    # Test that we can iterate over the paginated results
    # This validates the API contract that our pagination system depends on
    item_count = 0
    collected_ids: set[str] = set()  # Track unique IDs
    item: TextToCadResponse
    for item in iterator:
        item_count += 1
        # TextToCadResponse is a RootModel Union, so should have .root or .model_dump
        assert hasattr(item, "root") or hasattr(item, "model_dump"), (
            "Items should be valid TextToCadResponse models"
        )

        # Extract ID from the TextToCadResponse (which is a RootModel)
        if hasattr(item, "root"):
            # RootModel - get the underlying object
            underlying_obj = item.root  # type: ignore[attr-defined]
            item_id = str(underlying_obj.id)  # type: ignore[attr-defined]
        else:
            # Fallback - get id directly if possible
            item_id = str(getattr(item, "id", f"unknown_{item_count}"))

        # Verify this ID is unique
        assert item_id not in collected_ids, (
            f"Duplicate ID found: {item_id} (item #{item_count}). "
            f"Previously seen IDs: {sorted(collected_ids)}"
        )
        collected_ids.add(item_id)

        # Show which "page" this item likely came from (since limit=10)
        expected_page = ((item_count - 1) // 10) + 1
        print(f"Item #{item_count} (page ~{expected_page}): ID = {item_id}")

        # Don't iterate through all items in integration test, just verify it works
        if item_count >= 12:
            break

    # Verify we got 12 unique items
    assert item_count >= 12, f"Should have at least 12 items, got {item_count}"
    assert len(collected_ids) >= 12, (
        f"Should have at least 12 unique IDs, got {len(collected_ids)}"
    )
    assert len(collected_ids) == item_count, (
        f"All {item_count} items should have unique IDs, but got {len(collected_ids)} unique IDs"
    )

    print(
        f"✅ Successfully validated {item_count} items with {len(collected_ids)} unique IDs"
    )


@pytest.mark.asyncio
async def test_async_pagination_integration_text_to_cad():
    """Integration test for async pagination with real text-to-cad API calls."""
    # Create async client with real API token
    client = AsyncKittyCAD()

    # Call paginated endpoint - now returns AsyncPageIterator directly
    iterator = client.ml.list_text_to_cad_models_for_user(  # type: ignore[attr-defined]
        limit=10,  # Small page size to test pagination
    )

    # Verify we got an AsyncPageIterator
    from kittycad.models.text_to_cad_response import TextToCadResponse
    from kittycad.pagination import AsyncPageIterator

    assert isinstance(iterator, AsyncPageIterator), (
        f"Expected AsyncPageIterator, got {type(iterator)}"
    )

    # Test that we can iterate over the paginated results
    # This validates the API contract that our pagination system depends on
    item_count = 0
    collected_ids: set[str] = set()  # Track unique IDs
    item: TextToCadResponse
    async for item in iterator:
        item_count += 1
        # TextToCadResponse is a RootModel Union, so should have .root or .model_dump
        assert hasattr(item, "root") or hasattr(item, "model_dump"), (
            "Items should be valid TextToCadResponse models"
        )

        # Extract ID from the TextToCadResponse (which is a RootModel)
        if hasattr(item, "root"):
            # RootModel - get the underlying object
            underlying_obj = item.root  # type: ignore[attr-defined]
            item_id = str(underlying_obj.id)  # type: ignore[attr-defined]
        else:
            # Fallback - get id directly if possible
            item_id = str(getattr(item, "id", f"unknown_{item_count}"))

        # Verify this ID is unique
        assert item_id not in collected_ids, (
            f"Duplicate ID found: {item_id} (item #{item_count}). "
            f"Previously seen IDs: {sorted(collected_ids)}"
        )
        collected_ids.add(item_id)

        # Show which "page" this item likely came from (since limit=10)
        expected_page = ((item_count - 1) // 10) + 1
        print(f"Async Item #{item_count} (page ~{expected_page}): ID = {item_id}")

        # Don't iterate through all items in integration test, just verify it works
        if item_count >= 12:
            break

    # Verify we got 12 unique items
    assert item_count >= 12, f"Should have at least 12 items, got {item_count}"
    assert len(collected_ids) >= 12, (
        f"Should have at least 12 unique IDs, got {len(collected_ids)}"
    )
    assert len(collected_ids) == item_count, (
        f"All {item_count} items should have unique IDs, but got {len(collected_ids)} unique IDs"
    )

    print(
        f"✅ Async: Successfully validated {item_count} items with {len(collected_ids)} unique IDs"
    )

    # Clean up the async client
    await client.aclose()


def test_pagination_url_parameters_no_duplication():
    """Test that pagination URL parameters are not duplicated."""
    from unittest.mock import Mock, patch

    from kittycad import KittyCAD
    from kittycad.pagination import SyncPageIterator

    # Create a mock client to intercept HTTP requests
    client = KittyCAD()

    # Mock the HTTP client to capture URLs for each request
    captured_urls = []

    def mock_http_request(url, **kwargs):
        nonlocal captured_urls
        captured_urls.append(url)
        # Return a mock response that looks like a pagination response
        mock_response = Mock()
        mock_response.is_success = True
        # First request returns a page with next_page token, second returns final page
        if len(captured_urls) == 1:
            mock_response.content = b'{"items": [], "next_page": "page2_token"}'
            mock_response.json.return_value = {"items": [], "next_page": "page2_token"}
        else:
            mock_response.content = b'{"items": [], "next_page": null}'
            mock_response.json.return_value = {"items": [], "next_page": None}
        return mock_response

    # Get the iterator first with specific parameters
    from kittycad.models.created_at_sort_mode import CreatedAtSortMode

    iterator = client.ml.list_text_to_cad_models_for_user(
        limit=5, sort_by=CreatedAtSortMode.CREATED_AT_ASCENDING
    )
    assert isinstance(iterator, SyncPageIterator)

    # Patch the HTTP client to mock the request
    with patch.object(client, "get_http_client") as mock_get_client:
        mock_client = Mock()
        mock_client.get = mock_http_request
        mock_get_client.return_value = mock_client

        # Trigger pagination by fully consuming the iterator
        list(iterator)  # Consume the iterator to trigger HTTP requests

        # Verify we made exactly 2 requests (first page + second page)
        assert len(captured_urls) == 2, (
            f"Expected 2 HTTP requests, got {len(captured_urls)}"
        )

        # Check first URL (initial request)
        first_url = captured_urls[0]
        print(f"First request URL: {first_url}")

        # Verify first URL has correct parameters and no duplicates
        limit_count = first_url.count("limit=5")
        assert limit_count == 1, (
            f"Expected 'limit=5' to appear exactly once in first URL, but found {limit_count} occurrences in: {first_url}"
        )

        sort_by_count = first_url.count("sort_by=created_at_ascending")
        assert sort_by_count == 1, (
            f"Expected 'sort_by=created_at_ascending' to appear exactly once in first URL, but found {sort_by_count} occurrences in: {first_url}"
        )

        # First request should not have page_token
        page_token_count = first_url.count("page_token=")
        assert page_token_count == 0, (
            f"Expected no 'page_token' in first URL, but found {page_token_count} occurrences in: {first_url}"
        )

        # Check second URL (pagination request)
        second_url = captured_urls[1]
        print(f"Second request URL: {second_url}")

        # Verify second URL has page_token from first response
        page_token_count = second_url.count("page_token=page2_token")
        assert page_token_count == 1, (
            f"Expected 'page_token=page2_token' to appear exactly once in second URL, but found {page_token_count} occurrences in: {second_url}"
        )

        # Verify limit parameter is preserved in second request
        limit_count = second_url.count("limit=5")
        assert limit_count == 1, (
            f"Expected 'limit=5' to appear exactly once in second URL, but found {limit_count} occurrences in: {second_url}"
        )

        # Verify both URLs are properly formed
        for url in captured_urls:
            assert "user/text-to-cad" in url
            assert "limit=5" in url

        print("All pagination URLs verified - no parameter duplication detected")


def test_pagination_explicit_page_token():
    """Test that explicitly passed page_token is correctly used in initial request."""
    from unittest.mock import Mock, patch

    from kittycad import KittyCAD
    from kittycad.pagination import SyncPageIterator

    # Create a mock client to intercept HTTP requests
    client = KittyCAD()

    # Mock the HTTP client to capture the URL
    captured_url = None

    def mock_http_request(url, **kwargs):
        nonlocal captured_url
        captured_url = url
        # Return a mock response that looks like a final pagination page
        mock_response = Mock()
        mock_response.is_success = True
        mock_response.content = b'{"items": [], "next_page": null}'
        mock_response.json.return_value = {"items": [], "next_page": None}
        return mock_response

    # Get the iterator with an explicit page_token (user wants to start from specific page)
    iterator = client.ml.list_text_to_cad_models_for_user(
        limit=5, page_token="explicit_start_token"
    )
    assert isinstance(iterator, SyncPageIterator)

    # Patch the HTTP client to mock the request
    with patch.object(client, "get_http_client") as mock_get_client:
        mock_client = Mock()
        mock_client.get = mock_http_request
        mock_get_client.return_value = mock_client

        # Trigger the first page fetch by starting iteration
        result_items: list = list(iterator)

        # Verify the URL was captured
        assert captured_url is not None, "No HTTP request was made"
        print(f"URL with explicit page_token: {captured_url}")

        # Check that limit parameter appears exactly once
        limit_count = captured_url.count("limit=5")
        assert limit_count == 1, (
            f"Expected 'limit=5' to appear exactly once, but found {limit_count} occurrences in URL: {captured_url}"
        )

        # Check that the explicit page_token appears exactly once
        page_token_count = captured_url.count("page_token=explicit_start_token")
        assert page_token_count == 1, (
            f"Expected 'page_token=explicit_start_token' to appear exactly once, but found {page_token_count} occurrences in URL: {captured_url}"
        )

        # Verify the URL structure looks correct
        assert "user/text-to-cad" in captured_url
        assert "limit=5" in captured_url
        assert "page_token=explicit_start_token" in captured_url

        # Ensure we got the expected result (empty pages)
        assert len(result_items) == 0
        print(f"Explicit page_token correctly passed to API: {captured_url}")



from ..models.axis import Axis
from ..models.direction import Direction
from .base import KittyCadBaseModel


class AxisDirectionPair(KittyCadBaseModel):
    """An [`Axis`] paired with a [`Direction`]."""

    axis: Axis

    direction: Direction



from typing import List

from .base import KittyCadBaseModel


class Solid3dGetAllEdgeFaces(KittyCadBaseModel):
    """The response from the `Solid3dGetAllEdgeFaces` command."""

    faces: List[str]



"""Pagination support for KittyCAD API with OpenAI-style auto-iteration."""

from typing import Any, AsyncIterator, Callable, Dict, Iterator, Optional, Type, TypeVar

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class SyncPageIterator:
    """Synchronous iterator for paginated API responses.

    Provides OpenAI-style auto-pagination that handles page tokens automatically.
    """

    def __init__(
        self,
        page_fetcher: Callable[..., BaseModel],
        initial_kwargs: Dict[str, Any],
        item_type: Optional[Type[T]] = None,
    ):
        """Initialize the sync page iterator.

        Args:
            page_fetcher: Function to fetch a page (e.g., client.api_calls.list_api_calls)
            initial_kwargs: Initial arguments for the first request
            item_type: Type of items being paginated
        """
        self._page_fetcher = page_fetcher
        self._initial_kwargs = initial_kwargs
        self._item_type = item_type
        self._current_page_token: Optional[str] = None
        self._exhausted = False

    def __iter__(self) -> Iterator[T]:
        """Return iterator that yields individual items across all pages."""
        # Reset state for new iteration
        self._current_page_token = None
        self._exhausted = False

        kwargs = self._initial_kwargs.copy()

        while not self._exhausted:
            # Add page token if we have one
            if self._current_page_token:
                kwargs["page_token"] = self._current_page_token
            # Note: Don't remove page_token from kwargs if it exists in initial_kwargs
            # This allows users to explicitly start pagination from a specific token

            # Fetch the page
            page = self._page_fetcher(**kwargs)

            # Extract items and yield them
            items = getattr(page, "items", [])
            # Handle case where items might be None
            if items is not None:
                for item in items:
                    yield item

            # Check for next page
            next_page_token = getattr(page, "next_page", None)
            if next_page_token:
                self._current_page_token = next_page_token
            else:
                self._exhausted = True
                break


class AsyncPageIterator:
    """Asynchronous iterator for paginated API responses.

    Provides OpenAI-style async auto-pagination that handles page tokens automatically.
    """

    def __init__(
        self,
        page_fetcher: Callable[..., Any],  # Returns Awaitable[BaseModel]
        initial_kwargs: Dict[str, Any],
        item_type: Optional[Type[T]] = None,
    ):
        """Initialize the async page iterator.

        Args:
            page_fetcher: Async function to fetch a page (e.g., client.api_calls.list_api_calls)
            initial_kwargs: Initial arguments for the first request
            item_type: Type of items being paginated
        """
        self._page_fetcher = page_fetcher
        self._initial_kwargs = initial_kwargs
        self._item_type = item_type
        self._current_page_token: Optional[str] = None
        self._exhausted = False

    def __aiter__(self) -> AsyncIterator[T]:
        """Return async iterator that yields individual items across all pages."""
        # Reset state for new iteration
        self._current_page_token = None
        self._exhausted = False
        return self._async_iter()

    async def _async_iter(self) -> AsyncIterator[T]:
        """Internal async iterator implementation."""
        kwargs = self._initial_kwargs.copy()

        while not self._exhausted:
            # Add page token if we have one
            if self._current_page_token:
                kwargs["page_token"] = self._current_page_token
            # Note: Don't remove page_token from kwargs if it exists in initial_kwargs
            # This allows users to explicitly start pagination from a specific token

            # Fetch the page
            page = await self._page_fetcher(**kwargs)

            # Extract items and yield them
            items = getattr(page, "items", [])
            # Handle case where items might be None
            if items is not None:
                for item in items:
                    yield item

            # Check for next page
            next_page_token = getattr(page, "next_page", None)
            if next_page_token:
                self._current_page_token = next_page_token
            else:
                self._exhausted = True
                break


def create_sync_page_iterator(
    page_fetcher: Callable[..., BaseModel],
    kwargs: Dict[str, Any],
    item_type: Optional[Type[T]] = None,
) -> SyncPageIterator:
    """Create a synchronous page iterator.

    Args:
        page_fetcher: Function that fetches a page
        kwargs: Arguments to pass to the page fetcher
        item_type: Type of individual items

    Returns:
        SyncPageIterator that can be iterated to get all items
    """
    return SyncPageIterator(page_fetcher, kwargs, item_type)


def create_async_page_iterator(
    page_fetcher: Callable[..., Any],  # Returns Awaitable[BaseModel]
    kwargs: Dict[str, Any],
    item_type: Optional[Type[T]] = None,
) -> AsyncPageIterator:
    """Create an asynchronous page iterator.

    Args:
        page_fetcher: Async function that fetches a page
        kwargs: Arguments to pass to the page fetcher
        item_type: Type of individual items

    Returns:
        AsyncPageIterator that can be async iterated to get all items
    """
    return AsyncPageIterator(page_fetcher, kwargs, item_type)



import datetime
from typing import Optional

from ..models.block_reason import BlockReason
from ..models.uuid import Uuid
from .base import KittyCadBaseModel


class ExtendedUser(KittyCadBaseModel):
    """Extended user information.

    This is mostly used for internal purposes. It returns a mapping of the user's information, including that of our third party services we use for users: Stripe"""

    block: Optional[BlockReason] = None

    can_train_on_data: bool = False

    company: Optional[str] = None

    created_at: datetime.datetime

    deletion_scheduled: bool = False

    discord: Optional[str] = None

    email: Optional[str] = None

    email_verified: Optional[datetime.datetime] = None

    first_name: Optional[str] = None

    github: Optional[str] = None

    hubspot_contact_id: Optional[str] = None

    id: Uuid

    image: str

    is_onboarded: bool = False

    is_service_account: bool = False

    last_name: Optional[str] = None

    name: Optional[str] = None

    phone: str = ""

    stripe_id: Optional[str] = None

    updated_at: datetime.datetime



from typing import Optional

from ..models.address_details import AddressDetails
from .base import KittyCadBaseModel


class BillingInfo(KittyCadBaseModel):
    """The billing information for payments."""

    address: Optional[AddressDetails] = None

    name: Optional[str] = None

    phone: str = ""



from typing import List

from .base import KittyCadBaseModel


class Solid3dGetAllOppositeEdges(KittyCadBaseModel):
    """The response from the `Solid3dGetAllOppositeEdges` command."""

    edges: List[str]



from typing import Any, Dict, Optional, Union

from pydantic import RootModel

from ..models.ml_tool_result import MlToolResult
from .base import KittyCadBaseModel
from .reasoning_message import ReasoningMessage


class Delta(KittyCadBaseModel):
    """"""

    delta: str


class ToolOutput(KittyCadBaseModel):
    """"""

    result: MlToolResult


class Error(KittyCadBaseModel):
    """"""

    detail: str


class Info(KittyCadBaseModel):
    """"""

    text: str


reasoning = ReasoningMessage


class EndOfStream(KittyCadBaseModel):
    """"""

    whole_response: Optional[str] = None


class MlCopilotServerMessage0(KittyCadBaseModel):
    """Delta of the response, e.g. a chunk of text/tokens."""

    delta: Dict[str, Any]


class MlCopilotServerMessage1(KittyCadBaseModel):
    """Completed tool call result."""

    tool_output: Dict[str, Any]


class MlCopilotServerMessage2(KittyCadBaseModel):
    """Error sent by server."""

    error: Dict[str, Any]


class MlCopilotServerMessage3(KittyCadBaseModel):
    """Log / banner text."""

    info: Dict[str, Any]


class MlCopilotServerMessage4(KittyCadBaseModel):
    """Assistant reasoning / chain-of-thought (if you expose it)."""

    reasoning: ReasoningMessage


class MlCopilotServerMessage5(KittyCadBaseModel):
    """Marks the end of a streamed answer."""

    end_of_stream: Dict[str, Any]


MlCopilotServerMessage = RootModel[
    Union[
        Delta,
        ToolOutput,
        Error,
        Info,
        ReasoningMessage,
        EndOfStream,
        MlCopilotServerMessage0,
        MlCopilotServerMessage1,
        MlCopilotServerMessage2,
        MlCopilotServerMessage3,
        MlCopilotServerMessage4,
        MlCopilotServerMessage5,
    ]
]



from enum import Enum


class CameraDragInteractionType(str, Enum):
    """The type of camera drag interaction."""  # noqa: E501

    """# Camera pan"""  # noqa: E501

    PAN = "pan"

    """# Camera rotate (spherical camera revolve/orbit)"""  # noqa: E501

    ROTATE = "rotate"

    """# Camera rotate (trackball with 3 degrees of freedom)"""  # noqa: E501

    ROTATETRACKBALL = "rotatetrackball"

    """# Camera zoom (increase or decrease distance to reference point center)"""  # noqa: E501

    ZOOM = "zoom"

    def __str__(self) -> str:
        return str(self.value)



from typing import Optional

from ..models.source_range import SourceRange
from .base import KittyCadBaseModel


class SourceRangePrompt(KittyCadBaseModel):
    """A source range and prompt for a text to CAD iteration."""

    file: Optional[str] = None

    prompt: str

    range: SourceRange



from .base import KittyCadBaseModel
from .base64data import Base64Data


class DerEncodedKeyPair(KittyCadBaseModel):
    """The DER encoded key pair."""

    private_key: Base64Data

    public_cert: Base64Data



from .base import KittyCadBaseModel


class DeviceAuthConfirmParams(KittyCadBaseModel):
    """The request parameters to confirm the `user_code` for the OAuth 2.0 Device Authorization Grant."""

    user_code: str



from typing import Optional

from .base import KittyCadBaseModel


class OrgDetails(KittyCadBaseModel):
    """The user-modifiable parts of an organization."""

    allow_users_in_domain_to_auto_join: Optional[bool] = None

    billing_email: Optional[str] = None

    domain: Optional[str] = None

    image: Optional[str] = None

    name: Optional[str] = None

    phone: str = ""



import datetime
from typing import Literal, Optional, Union

from pydantic import Field, RootModel
from typing_extensions import Annotated

from ..models.modeling_app_event_type import ModelingAppEventType
from .base import KittyCadBaseModel


class OptionModelingAppEvent(KittyCadBaseModel):
    """An event related to modeling app files"""

    attachment_uri: Optional[str] = None

    created_at: datetime.datetime

    event_type: ModelingAppEventType

    last_compiled_at: Optional[datetime.datetime] = None

    project_description: Optional[str] = None

    project_name: str

    source_id: str

    type: Literal["modeling_app_event"] = "modeling_app_event"

    user_id: str


Event = RootModel[
    Annotated[Union[OptionModelingAppEvent,], Field(discriminator="type")]
]



import datetime
from typing import Dict, Optional

from ..models.billing_info import BillingInfo
from ..models.card_details import CardDetails
from ..models.payment_method_type import PaymentMethodType
from .base import KittyCadBaseModel


class PaymentMethod(KittyCadBaseModel):
    """A payment method."""

    billing_info: BillingInfo

    card: Optional[CardDetails] = None

    created_at: datetime.datetime

    id: Optional[str] = None

    metadata: Dict[str, str] = {}

    type: PaymentMethodType



from typing import List

from ..models.export_file import ExportFile
from .base import KittyCadBaseModel


class Export2d(KittyCadBaseModel):
    """The response from the `Export2d` endpoint."""

    files: List[ExportFile]



from typing import List

from ..models.extrusion_face_info import ExtrusionFaceInfo
from .base import KittyCadBaseModel


class Solid3dGetExtrusionFaceInfo(KittyCadBaseModel):
    """Extrusion face info struct (useful for maintaining mappings between source path segment ids and extrusion faces)"""

    faces: List[ExtrusionFaceInfo]



from typing import Optional

from ..models.modeling_app_organization_subscription_tier import (
    ModelingAppOrganizationSubscriptionTier,
)
from .base import KittyCadBaseModel


class ZooProductSubscriptionsOrgRequest(KittyCadBaseModel):
    """A struct of Zoo product subscriptions an organization can request."""

    modeling_app: ModelingAppOrganizationSubscriptionTier = "team"  # type: ignore

    pay_annually: Optional[bool] = None



#!/usr/bin/env python3
"""
Example usage of the new exception-based KittyCAD Python SDK.

This demonstrates the clean, idiomatic interface where errors are handled
via exceptions rather than return values.
"""

from .. import KittyCAD, KittyCADAPIError, KittyCADClientError, KittyCADServerError


def test_example_successful_call():
    """Example of making a successful API call."""
    try:
        # Clean, simple interface using environment variable
        client = KittyCAD()  # Uses KITTYCAD_API_TOKEN environment variable
        result = client.meta.ping()
        # Either success or exception is acceptable for this test
        assert result is not None or result is None

    except KittyCADAPIError:
        # All API errors are caught here with rich context
        pass  # Expected when no valid token


def test_error_handling():
    """Example of handling different types of API errors."""
    try:
        # This will likely fail with authentication error
        client = KittyCAD()  # Uses KITTYCAD_API_TOKEN environment variable
        user = client.users.get_user_self()
        # Either success or exception is acceptable
        assert user is not None or user is None

    except KittyCADClientError as e:
        # Handle client errors (4xx) - usually fixable by user
        assert e.status_code >= 400 and e.status_code < 500

    except KittyCADServerError as e:
        # Handle server errors (5xx) - usually temporary
        assert e.status_code >= 500

    except KittyCADAPIError:
        # Catch-all for any other API errors
        pass


def test_interface_design():
    """Test that the new interface is properly designed."""
    # Test that KittyCAD class can be instantiated
    client = KittyCAD()
    assert hasattr(client, "users")
    assert hasattr(client, "meta")

    # Test that exception types exist
    assert KittyCADAPIError is not None
    assert KittyCADClientError is not None
    assert KittyCADServerError is not None



#!/usr/bin/env python3
"""Comprehensive tests for pagination behavior requirements."""

from typing import Iterator, List, Optional
from unittest.mock import AsyncMock, Mock, call

import pytest
from pydantic import BaseModel

from kittycad.pagination import AsyncPageIterator, SyncPageIterator


# Mock data models for testing
class MockItem(BaseModel):
    id: str
    name: str


class MockPage(BaseModel):
    items: List[MockItem]
    next_page: Optional[str] = None


class TestPaginationFirstCall:
    """Test first call pagination behavior."""

    def test_first_call_without_page_token_sends_scan_params_and_limit(self):
        """Test first call without page_token sends scan params and limit."""
        # Mock single page response
        items = [MockItem(id="1", name="Item 1")]
        page = MockPage(items=items, next_page=None)
        mock_fetcher = Mock(return_value=page)

        # Create iterator with scan params
        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={"limit": 10, "sort_by": "created_at", "filter": "active"},
            item_type=MockItem,
        )

        # Consume iterator
        list(iterator)

        # Verify first call includes all scan params and limit
        mock_fetcher.assert_called_once_with(
            limit=10, sort_by="created_at", filter="active"
        )
        # Should NOT include page_token in first call
        args, kwargs = mock_fetcher.call_args
        assert "page_token" not in kwargs

    def test_first_call_with_explicit_page_token_still_sends_scan_params(self):
        """Test first call with explicit page_token still includes scan params."""
        # Mock single page response
        items = [MockItem(id="1", name="Item 1")]
        page = MockPage(items=items, next_page=None)
        mock_fetcher = Mock(return_value=page)

        # Create iterator with explicit page_token in initial_kwargs
        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={
                "limit": 10,
                "sort_by": "created_at",
                "page_token": "explicit-token",
            },
            item_type=MockItem,
        )

        # Consume iterator
        list(iterator)

        # Verify first call includes page_token and scan params
        mock_fetcher.assert_called_once_with(
            limit=10, sort_by="created_at", page_token="explicit-token"
        )


class TestPaginationContinuation:
    """Test pagination continuation behavior."""

    def test_continuations_send_all_initial_params_plus_page_token(self):
        """Test continuations send all initial params plus page_token (current implementation)."""
        # Create multi-page mock data
        page1_items = [MockItem(id="1", name="Item 1")]
        page2_items = [MockItem(id="2", name="Item 2")]
        page3_items = [MockItem(id="3", name="Item 3")]

        page1 = MockPage(items=page1_items, next_page="token2")
        page2 = MockPage(items=page2_items, next_page="token3")
        page3 = MockPage(items=page3_items, next_page=None)

        def mock_fetch_page(**kwargs):
            if "page_token" not in kwargs:
                return page1
            elif kwargs["page_token"] == "token2":
                return page2
            elif kwargs["page_token"] == "token3":
                return page3
            else:
                raise ValueError(f"Unexpected page token: {kwargs['page_token']}")

        mock_fetcher = Mock(side_effect=mock_fetch_page)

        # Create iterator with scan params
        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={
                "limit": 2,
                "sort_by": "created_at",
                "filter": "active",
                "direction": "asc",
            },
            item_type=MockItem,
        )

        # Consume iterator
        result_items: list[MockItem] = list(iterator)

        # Verify we got all items
        assert len(result_items) == 3
        assert [item.id for item in result_items] == ["1", "2", "3"]

        # Verify call pattern
        assert mock_fetcher.call_count == 3

        # First call: all scan params, no page_token
        first_call = mock_fetcher.call_args_list[0]
        expected_first = call(
            limit=2, sort_by="created_at", filter="active", direction="asc"
        )
        assert first_call == expected_first

        # Current implementation: subsequent calls include ALL initial params + page_token
        second_call = mock_fetcher.call_args_list[1]
        expected_second = call(
            limit=2,
            sort_by="created_at",
            filter="active",
            direction="asc",
            page_token="token2",
        )
        assert second_call == expected_second

        third_call = mock_fetcher.call_args_list[2]
        expected_third = call(
            limit=2,
            sort_by="created_at",
            filter="active",
            direction="asc",
            page_token="token3",
        )
        assert third_call == expected_third

    def test_continuations_preserve_limit_if_provided(self):
        """Test continuations preserve limit parameter and all initial kwargs."""
        # Create multi-page mock data
        page1 = MockPage(items=[MockItem(id="1", name="Item 1")], next_page="token2")
        page2 = MockPage(items=[MockItem(id="2", name="Item 2")], next_page=None)

        def mock_fetch_page(**kwargs):
            if "page_token" not in kwargs:
                return page1
            elif kwargs["page_token"] == "token2":
                return page2
            else:
                raise ValueError(f"Unexpected page token: {kwargs['page_token']}")

        mock_fetcher = Mock(side_effect=mock_fetch_page)

        # Create iterator with custom limit
        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={"limit": 50, "sort_by": "created_at"},
            item_type=MockItem,
        )

        # Consume iterator
        list(iterator)

        # Verify both calls preserved the limit and all params
        assert mock_fetcher.call_count == 2

        first_call = mock_fetcher.call_args_list[0]
        assert first_call == call(limit=50, sort_by="created_at")

        second_call = mock_fetcher.call_args_list[1]
        assert second_call == call(limit=50, sort_by="created_at", page_token="token2")
        # Current implementation: scan params ARE included in continuation calls
        assert second_call[1]["sort_by"] == "created_at"

    def test_continuations_without_limit_omit_limit(self):
        """Test continuations without initial limit maintain all initial params."""
        # Create multi-page mock data
        page1 = MockPage(items=[MockItem(id="1", name="Item 1")], next_page="token2")
        page2 = MockPage(items=[MockItem(id="2", name="Item 2")], next_page=None)

        def mock_fetch_page(**kwargs):
            if "page_token" not in kwargs:
                return page1
            elif kwargs["page_token"] == "token2":
                return page2
            else:
                raise ValueError(f"Unexpected page token: {kwargs['page_token']}")

        mock_fetcher = Mock(side_effect=mock_fetch_page)

        # Create iterator without limit
        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={"sort_by": "created_at"},
            item_type=MockItem,
        )

        # Consume iterator
        list(iterator)

        # Verify call pattern
        assert mock_fetcher.call_count == 2

        # First call: scan params, no limit
        first_call = mock_fetcher.call_args_list[0]
        assert first_call == call(sort_by="created_at")
        assert "limit" not in first_call[1]

        # Second call: all initial params + page_token, still no limit
        second_call = mock_fetcher.call_args_list[1]
        assert second_call == call(sort_by="created_at", page_token="token2")
        assert "limit" not in second_call[1]
        # Current implementation: scan params ARE included in continuation calls
        assert second_call[1]["sort_by"] == "created_at"


class TestPaginationMethods:
    """Test pagination iterator methods."""

    def test_has_next_page_method(self):
        """Test .has_next_page() works correctly."""
        # Note: Current implementation doesn't have has_next_page method
        # This test documents expected behavior for future implementation
        iterator = SyncPageIterator(
            page_fetcher=Mock(),
            initial_kwargs={"limit": 10},
            item_type=MockItem,
        )

        # Should have method for checking if there are more pages
        # assert hasattr(iterator, 'has_next_page')

        # For now, we can test internal state tracking
        assert not iterator._exhausted
        assert iterator._current_page_token is None

    def test_next_page_info_method(self):
        """Test .next_page_info() works correctly."""
        # Note: Current implementation doesn't have next_page_info method
        # This test documents expected behavior for future implementation
        iterator = SyncPageIterator(
            page_fetcher=Mock(),
            initial_kwargs={"limit": 10},
            item_type=MockItem,
        )

        # Should have method for getting next page info
        # assert hasattr(iterator, 'next_page_info')

        # For now, we can test internal state tracking
        assert iterator._current_page_token is None

    def test_get_next_page_method(self):
        """Test .get_next_page() works correctly."""
        # Note: Current implementation doesn't have get_next_page method
        # This test documents expected behavior for future implementation
        SyncPageIterator(
            page_fetcher=Mock(),
            initial_kwargs={"limit": 10},
            item_type=MockItem,
        )

        # Should have method for manually getting next page
        # assert hasattr(iterator, 'get_next_page')

        # Current implementation handles pagination automatically
        # through iteration, which is the expected behavior


class TestPaginationIteration:
    """Test pagination iteration behavior."""

    def test_item_iteration_auto_fetches_all_pages(self):
        """Test item iteration auto-fetches all pages, yields in order, stops at end."""
        # Create mock data for 3 pages
        page1_items = [MockItem(id="1", name="Item 1"), MockItem(id="2", name="Item 2")]
        page2_items = [MockItem(id="3", name="Item 3"), MockItem(id="4", name="Item 4")]
        page3_items = [MockItem(id="5", name="Item 5")]

        page1 = MockPage(items=page1_items, next_page="token2")
        page2 = MockPage(items=page2_items, next_page="token3")
        page3 = MockPage(items=page3_items, next_page=None)

        def mock_fetch_page(**kwargs):
            if "page_token" not in kwargs:
                return page1
            elif kwargs["page_token"] == "token2":
                return page2
            elif kwargs["page_token"] == "token3":
                return page3
            else:
                raise ValueError(f"Unexpected page token: {kwargs['page_token']}")

        mock_fetcher = Mock(side_effect=mock_fetch_page)

        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={"limit": 2},
            item_type=MockItem,
        )

        # Iterate and collect items in order
        result_items: list[MockItem] = []
        item: MockItem
        for item in iterator:
            result_items.append(item)

        # Verify correct order and all items fetched
        assert len(result_items) == 5
        assert [item.id for item in result_items] == ["1", "2", "3", "4", "5"]
        assert [item.name for item in result_items] == [
            "Item 1",
            "Item 2",
            "Item 3",
            "Item 4",
            "Item 5",
        ]

        # Verify all pages were fetched
        assert mock_fetcher.call_count == 3

        # Verify iteration stops at end (no more calls after exhaustion)
        # Try iterating again - should reset and start over
        result_items2: list[MockItem] = list(iterator)
        assert len(result_items2) == 5
        assert mock_fetcher.call_count == 6  # Another 3 calls for second iteration

    def test_iteration_stops_at_end_correctly(self):
        """Test iteration stops correctly when no more pages."""
        # Single page with no next_page
        page = MockPage(
            items=[MockItem(id="1", name="Item 1"), MockItem(id="2", name="Item 2")],
            next_page=None,
        )
        mock_fetcher = Mock(return_value=page)

        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={"limit": 10},
            item_type=MockItem,
        )

        # Consume iterator
        result_items: list[MockItem] = list(iterator)

        # Verify correct items and single page fetch
        assert len(result_items) == 2
        assert mock_fetcher.call_count == 1

        # Verify iterator is exhausted
        assert iterator._exhausted

    def test_empty_pages_handled_correctly(self):
        """Test empty pages are handled correctly."""
        # Page with empty items list
        empty_page = MockPage(items=[], next_page=None)
        mock_fetcher = Mock(return_value=empty_page)

        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={"limit": 10},
            item_type=MockItem,
        )

        # Consume iterator
        result_items: list[MockItem] = list(iterator)

        # Should handle empty page gracefully
        assert len(result_items) == 0
        assert mock_fetcher.call_count == 1


class TestPaginationMemory:
    """Test pagination memory efficiency."""

    def test_large_sequence_iteration_keeps_o1_memory(self):
        """Test large sequence iteration keeps O(1) extra memory."""
        # Create iterator for simulated large dataset
        large_page_count = 100
        items_per_page = 100

        def mock_fetch_large_page(**kwargs):
            """Simulate fetching a large page."""
            if "page_token" not in kwargs:
                page_num = 1
            else:
                page_num = int(kwargs["page_token"].split("-")[1])

            # Create items for this page
            start_id = (page_num - 1) * items_per_page + 1
            items = [
                MockItem(id=str(i), name=f"Item {i}")
                for i in range(start_id, start_id + items_per_page)
            ]

            # Determine next page token
            next_page = f"page-{page_num + 1}" if page_num < large_page_count else None

            return MockPage(items=items, next_page=next_page)

        mock_fetcher = Mock(side_effect=mock_fetch_large_page)

        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={"limit": items_per_page},
            item_type=MockItem,
        )

        # Track memory usage by counting how many items we keep in memory
        processed_count = 0
        max_items_in_memory = 0

        # Process items one at a time (simulating real usage)
        item: MockItem
        for item in iterator:
            processed_count += 1

            # In real iteration, we should only have one item in memory at a time
            # The iterator should not accumulate all items
            max_items_in_memory = max(max_items_in_memory, 1)

            # Process a few pages to verify behavior
            if processed_count >= 250:  # 2.5 pages worth
                break

        # Verify we processed items without accumulating them
        assert processed_count == 250
        assert max_items_in_memory == 1  # O(1) memory usage

        # Verify pages were fetched on demand (should be 3 pages for 250 items)
        assert mock_fetcher.call_count == 3

    def test_iterator_does_not_preload_all_pages(self):
        """Test iterator does not preload all pages into memory."""
        call_count = 0

        def mock_fetch_on_demand(**kwargs):
            """Track when pages are fetched."""
            nonlocal call_count
            call_count += 1

            if "page_token" not in kwargs:
                return MockPage(
                    items=[MockItem(id="1", name="Page 1 Item")], next_page="token2"
                )
            elif kwargs["page_token"] == "token2":
                return MockPage(
                    items=[MockItem(id="2", name="Page 2 Item")], next_page="token3"
                )
            else:
                return MockPage(
                    items=[MockItem(id="3", name="Page 3 Item")], next_page=None
                )

        mock_fetcher = Mock(side_effect=mock_fetch_on_demand)

        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={"limit": 1},
            item_type=MockItem,
        )

        # Start iteration but don't consume all items
        iter_obj: Iterator[MockItem] = iter(iterator)

        # Get first item - should only fetch first page
        first_item = next(iter_obj)
        assert first_item.id == "1"
        assert call_count == 1  # Only first page fetched

        # Get second item - should fetch second page
        second_item = next(iter_obj)
        assert second_item.id == "2"
        assert call_count == 2  # Second page fetched on demand

        # Verify we haven't preloaded the third page yet
        # Only fetch it when we actually need it
        third_item = next(iter_obj)
        assert third_item.id == "3"
        assert call_count == 3  # Third page fetched on demand


class TestPaginationScanParams:
    """Test pagination scan parameter behavior."""

    def test_changing_scan_params_mid_scan_ignored(self):
        """Test changing scan params mid-scan is ignored."""
        # Create multi-page mock data
        page1 = MockPage(items=[MockItem(id="1", name="Item 1")], next_page="token2")
        page2 = MockPage(items=[MockItem(id="2", name="Item 2")], next_page=None)

        calls_received = []

        def mock_fetch_page(**kwargs):
            calls_received.append(kwargs.copy())
            if "page_token" not in kwargs:
                return page1
            else:
                return page2

        mock_fetcher = Mock(side_effect=mock_fetch_page)

        iterator = SyncPageIterator(
            page_fetcher=mock_fetcher,
            initial_kwargs={"limit": 1, "sort_by": "created_at"},
            item_type=MockItem,
        )

        # Start iteration
        iter_obj: Iterator[MockItem] = iter(iterator)
        next(iter_obj)

        # Try to modify initial_kwargs (this should not affect ongoing iteration)
        iterator._initial_kwargs["sort_by"] = "modified_at"
        iterator._initial_kwargs["new_param"] = "new_value"

        # Continue iteration - should not use modified params
        next(iter_obj)

        # Verify the calls received
        assert len(calls_received) == 2

        # First call: original scan params
        assert calls_received[0] == {"limit": 1, "sort_by": "created_at"}

        # Second call: all initial params + page_token (modified params ignored)
        assert calls_received[1] == {
            "limit": 1,
            "sort_by": "created_at",
            "page_token": "token2",
        }
        # The original sort_by value should be preserved, modifications ignored
        assert calls_received[1]["sort_by"] == "created_at"  # Original value
        assert "new_param" not in calls_received[1]  # New param not included


class TestPaginationAsync:
    """Test async pagination behaves the same as sync."""

    @pytest.mark.asyncio
    async def test_async_for_behaves_same_as_sync(self):
        """Test async for behaves the same as sync iteration."""
        # Create same test data for both sync and async
        page1_items = [MockItem(id="1", name="Item 1"), MockItem(id="2", name="Item 2")]
        page2_items = [MockItem(id="3", name="Item 3")]

        page1 = MockPage(items=page1_items, next_page="token2")
        page2 = MockPage(items=page2_items, next_page=None)

        # Sync version
        def sync_fetch_page(**kwargs):
            if "page_token" not in kwargs:
                return page1
            else:
                return page2

        sync_mock_fetcher = Mock(side_effect=sync_fetch_page)
        sync_iterator = SyncPageIterator(
            page_fetcher=sync_mock_fetcher,
            initial_kwargs={"limit": 2, "sort_by": "created_at"},
            item_type=MockItem,
        )

        sync_results: list[MockItem] = list(sync_iterator)

        # Async version
        async def async_fetch_page(**kwargs):
            if "page_token" not in kwargs:
                return page1
            else:
                return page2

        async_mock_fetcher = AsyncMock(side_effect=async_fetch_page)
        async_iterator = AsyncPageIterator(
            page_fetcher=async_mock_fetcher,
            initial_kwargs={"limit": 2, "sort_by": "created_at"},
            item_type=MockItem,
        )

        async_results: list[MockItem] = []
        item: MockItem
        async for item in async_iterator:
            async_results.append(item)

        # Both should yield same results
        assert len(sync_results) == len(async_results) == 3
        for sync_item, async_item in zip(sync_results, async_results):
            assert sync_item.id == async_item.id
            assert sync_item.name == async_item.name

        # Both should make same API calls
        assert sync_mock_fetcher.call_count == async_mock_fetcher.call_count == 2

        # Verify call patterns are identical
        sync_calls = [call[1] for call in sync_mock_fetcher.call_args_list]
        async_calls = [call[1] for call in async_mock_fetcher.call_args_list]
        assert sync_calls == async_calls

    @pytest.mark.asyncio
    async def test_async_pagination_memory_efficiency(self):
        """Test async pagination has same memory efficiency as sync."""
        call_count = 0

        async def async_fetch_on_demand(**kwargs):
            nonlocal call_count
            call_count += 1

            if "page_token" not in kwargs:
                return MockPage(
                    items=[MockItem(id="1", name="Page 1 Item")], next_page="token2"
                )
            elif kwargs["page_token"] == "token2":
                return MockPage(
                    items=[MockItem(id="2", name="Page 2 Item")], next_page=None
                )
            else:
                raise ValueError("Unexpected token")

        async_mock_fetcher = AsyncMock(side_effect=async_fetch_on_demand)

        iterator = AsyncPageIterator(
            page_fetcher=async_mock_fetcher,
            initial_kwargs={"limit": 1},
            item_type=MockItem,
        )

        # Consume items using async for instead of aiter/anext (Python 3.9 compatibility)
        items: list[MockItem] = []
        item: MockItem
        async for item in iterator:
            items.append(item)
            # Check call count after each item
            if len(items) == 1:
                assert call_count == 1  # Only first page fetched
            if len(items) >= 2:  # Stop after getting 2 items
                break

        assert items[0].id == "1"
        assert items[1].id == "2"
        assert call_count == 2  # Second page fetched on demand


if __name__ == "__main__":
    pytest.main([__file__])



from enum import Enum


class Direction(str, Enum):
    """Specifies the sign of a co-ordinate axis."""  # noqa: E501

    """# Increasing numbers."""  # noqa: E501

    POSITIVE = "positive"

    """# Decreasing numbers."""  # noqa: E501

    NEGATIVE = "negative"

    def __str__(self) -> str:
        return str(self.value)



#!/usr/bin/env python3
"""Tests for Unset/UNSET behavior and Pydantic model functionality."""

import json
from enum import Enum
from typing import Optional, Union

import pytest
from pydantic import BaseModel, ValidationError

from kittycad.models import ApiCallStatus, FileExportFormat
from kittycad.types import UNSET, Unset


class TestUnsetBehavior:
    """Test Unset/UNSET parameter behavior."""

    def test_unset_is_falsy(self):
        """Test UNSET evaluates to False in boolean context."""
        assert not UNSET
        assert bool(UNSET) is False

        # Test isinstance behavior
        assert isinstance(UNSET, Unset)

        # Test type consistency
        unset_instance = Unset()
        assert not unset_instance
        assert bool(unset_instance) is False

    def test_unset_type_annotations(self):
        """Test Unset appears in type annotations where expected."""
        # Test that Unset type exists and can be used in type annotations
        # This documents the expected type annotation patterns

        # Example of how Unset would be used in function signatures
        def example_function(
            required_field: str,
            optional_with_unset: Union[str, Unset] = UNSET,
            optional_with_none: Optional[str] = None,
        ) -> dict:
            return {
                "required": required_field,
                "unset": optional_with_unset,
                "none": optional_with_none,
            }

        # Test function with various Unset combinations
        result1 = example_function("test")
        assert result1["required"] == "test"
        assert result1["unset"] is UNSET
        assert result1["none"] is None

        result2 = example_function("test", "provided", "also_provided")
        assert result2["unset"] == "provided"
        assert result2["none"] == "also_provided"

    def test_unset_vs_none_serialization(self):
        """Test UNSET vs None serialization behavior."""
        # Test serialization behavior without using Pydantic models
        # since Unset doesn't have Pydantic schema support yet

        def serialize_value(value):
            """Simulate how values should be serialized."""
            if value is UNSET:
                return None  # Omit from output
            elif value is None:
                return "null"  # Serialize as null
            else:
                return str(value)  # Regular serialization

        # Test different values
        assert serialize_value(UNSET) is None  # Should be omitted
        assert serialize_value(None) == "null"  # Should be serialized as null
        assert serialize_value("test") == "test"  # Regular value

        # Test that we can distinguish between UNSET and None
        assert UNSET is not None
        assert UNSET is not None
        assert None is not UNSET

    def test_unset_field_omitted_from_json(self):
        """Test when param is UNSET, field is omitted from JSON."""
        # Test serialization logic without Pydantic models

        def build_request_dict(
            name: str, optional_param: Union[str, Unset] = UNSET
        ) -> dict:
            """Simulate building request dict, omitting UNSET fields."""
            result = {"name": name}
            if optional_param is not UNSET:
                result["optional_param"] = str(optional_param)
            return result

        # Model with UNSET field
        request_with_unset = build_request_dict("test")

        # Should only include name, not optional_param
        assert "name" in request_with_unset
        assert "optional_param" not in request_with_unset
        assert request_with_unset == {"name": "test"}

        # Model with provided field
        request_with_value = build_request_dict("test", "provided")
        assert "name" in request_with_value
        assert "optional_param" in request_with_value
        assert request_with_value == {"name": "test", "optional_param": "provided"}

    def test_none_field_serialized_as_null(self):
        """Test when param is None, field serialized as null."""
        # Test serialization logic without Pydantic models

        def build_request_with_none(
            name: str, nullable_param: Optional[str] = None
        ) -> dict:
            """Build request dict, including None fields."""
            return {"name": name, "nullable_param": nullable_param}

        # Model with None field
        request_with_none = build_request_with_none("test", None)

        # Should include both fields, with nullable_param as null
        assert "name" in request_with_none
        assert "nullable_param" in request_with_none
        assert request_with_none["nullable_param"] is None
        assert request_with_none == {"name": "test", "nullable_param": None}

    def test_required_properties_reject_unset(self):
        """Test required properties must reject UNSET."""
        # Test validation logic without Pydantic models

        def validate_required_field(required_field: str) -> dict:
            """Validate that required field is provided and not UNSET."""
            if required_field is UNSET:
                raise ValueError("Required field cannot be UNSET")
            if not isinstance(required_field, str):
                raise TypeError("Required field must be a string")
            return {"required_field": required_field}

        # Should succeed with valid value
        result = validate_required_field("test")
        assert result == {"required_field": "test"}

        # Should raise error when required field is UNSET
        with pytest.raises(ValueError, match="Required field cannot be UNSET"):
            validate_required_field(UNSET)  # type: ignore

        # Should raise error for wrong type
        with pytest.raises(TypeError):
            validate_required_field(123)  # type: ignore

    def test_unset_with_default_values(self):
        """Test UNSET behavior with default values."""
        # Test default value behavior without Pydantic models

        def build_with_defaults(
            field_with_string_default: str = "default_value",
            field_with_unset_default: Union[str, Unset] = UNSET,
            field_with_none_default: Optional[str] = None,
        ) -> dict[str, Union[str, None]]:
            """Build dict with defaults, excluding UNSET fields."""
            result: dict[str, Union[str, None]] = {
                "field_with_string_default": field_with_string_default
            }

            if field_with_unset_default is not UNSET:
                result["field_with_unset_default"] = str(field_with_unset_default)

            result["field_with_none_default"] = field_with_none_default
            return result

        # Create with defaults
        defaults_result = build_with_defaults()

        assert defaults_result["field_with_string_default"] == "default_value"
        assert "field_with_unset_default" not in defaults_result  # Should be excluded
        assert "field_with_none_default" in defaults_result  # None is included
        assert defaults_result["field_with_none_default"] is None

        # Create with explicit values
        explicit_result = build_with_defaults("custom", "provided", "also_provided")
        assert explicit_result["field_with_string_default"] == "custom"
        assert explicit_result["field_with_unset_default"] == "provided"
        assert explicit_result["field_with_none_default"] == "also_provided"


class TestPydanticModels:
    """Test Pydantic v2 model functionality."""

    def test_model_dump_matches_server_schema(self):
        """Test .model_dump() produces JSON matching server schema."""
        # Use a real model from the codebase
        from kittycad.models import (
            Axis,
            AxisDirectionPair,
            ConversionParams,
            Direction,
            System,
            UnitLength,
        )
        from kittycad.models.input_format3d import InputFormat3d, OptionStl
        from kittycad.models.output_format3d import (
            OptionObj as OutputOptionObj,
            OutputFormat3d,
        )

        # Create a complex model
        params = ConversionParams(
            src_format=InputFormat3d(
                OptionStl(
                    coords=System(
                        forward=AxisDirectionPair(
                            axis=Axis.Y,
                            direction=Direction.NEGATIVE,
                        ),
                        up=AxisDirectionPair(
                            axis=Axis.Z,
                            direction=Direction.POSITIVE,
                        ),
                    ),
                    units=UnitLength.MM,
                )
            ),
            output_format=OutputFormat3d(
                OutputOptionObj(
                    coords=System(
                        forward=AxisDirectionPair(
                            axis=Axis.Y,
                            direction=Direction.POSITIVE,
                        ),
                        up=AxisDirectionPair(
                            axis=Axis.Z,
                            direction=Direction.POSITIVE,
                        ),
                    ),
                    units=UnitLength.MM,
                )
            ),
        )

        # Test model_dump produces valid JSON
        dumped = params.model_dump()
        assert isinstance(dumped, dict)

        # Should be JSON serializable
        json_str = json.dumps(dumped)
        assert isinstance(json_str, str)

        # Should be able to parse back
        parsed = json.loads(json_str)
        assert isinstance(parsed, dict)

    def test_model_validate_builds_correct_object(self):
        """Test .model_validate(server_json) builds correct object."""
        # Create server-like JSON response
        server_json = {
            "status": "completed",
            "id": "test-id-123",
            "created_at": "2023-01-01T00:00:00Z",
            "updated_at": "2023-01-01T00:01:00Z",
        }

        # Import a real model for testing
        from kittycad.models import ApiCallWithPrice

        # Test model_validate can build object from server JSON
        try:
            # This may fail if ApiCallWithPrice requires different fields
            # But demonstrates the pattern
            model = ApiCallWithPrice.model_validate(server_json)
            assert model.id == "test-id-123"
        except (ValidationError, AttributeError):
            # If the exact fields don't match, test the pattern with a simpler approach
            pass

        # Test round-trip: dump -> validate with a simpler model
        # Use ApiCallStatus enum which we know works

        # Test enum round-trip
        status = ApiCallStatus.COMPLETED
        status_dict = {"status": status.value}

        # Test that we can validate enum values
        assert ApiCallStatus(status_dict["status"]) == ApiCallStatus.COMPLETED

    def test_enums_use_python_enum_and_serialize_to_strings(self):
        """Test enums use Python Enum, values serialize to string constants."""
        # Test that status enums are proper Python enums
        assert isinstance(ApiCallStatus.COMPLETED, ApiCallStatus)
        assert issubclass(ApiCallStatus, Enum)

        # Test enum values serialize to strings
        assert ApiCallStatus.COMPLETED.value == "completed"
        assert ApiCallStatus.IN_PROGRESS.value == "in_progress"
        assert ApiCallStatus.FAILED.value == "failed"

        # Test in JSON serialization context using a simple dict
        test_data = {"status": ApiCallStatus.COMPLETED}

        # When serializing, enum should become string
        import json

        json_str = json.dumps(
            test_data, default=lambda x: x.value if hasattr(x, "value") else str(x)
        )
        parsed = json.loads(json_str)
        assert parsed["status"] == "completed"  # String, not enum object

        # Test another enum
        assert isinstance(FileExportFormat.OBJ, FileExportFormat)
        assert FileExportFormat.OBJ.value == "obj"
        assert FileExportFormat.STL.value == "stl"

    def test_pydantic_model_serialization_consistency(self):
        """Test Pydantic models serialize consistently with simple models."""
        # Test with enum serialization directly
        status = ApiCallStatus.IN_PROGRESS

        # Test that enum has correct value
        assert status.value == "in_progress"

        # Test JSON serialization of enum values
        import json

        data = {"status": status.value, "id": "test-123"}
        json_str = json.dumps(data)
        assert isinstance(json_str, str)

        # Should be parseable JSON
        parsed = json.loads(json_str)
        assert parsed["id"] == "test-123"
        assert parsed["status"] == "in_progress"

        # Test that we can reconstruct enum from string
        reconstructed_status = ApiCallStatus(parsed["status"])
        assert reconstructed_status == ApiCallStatus.IN_PROGRESS

    def test_pydantic_validation_errors(self):
        """Test Pydantic validation errors are informative."""
        # Test with enum validation directly
        with pytest.raises(ValueError) as exc_info:
            ApiCallStatus("invalid_status")  # Invalid enum value

        # Should raise ValueError for invalid enum
        assert (
            "invalid_status" in str(exc_info.value).lower()
            or "not a valid" in str(exc_info.value).lower()
        )


class TestTypingStubs:
    """Test typing stubs and type annotations."""

    def test_py_typed_marker_exists(self):
        """Test py.typed exists for type checker discovery."""
        import os

        import kittycad

        # Check if py.typed file exists in the package
        package_dir = os.path.dirname(kittycad.__file__)
        py_typed_path = os.path.join(package_dir, "py.typed")

        # Verify the py.typed marker file exists
        assert os.path.exists(py_typed_path), (
            "py.typed marker file should exist for type checking"
        )

    def test_overloaded_function_stubs(self):
        """Test stubs for overloaded functions are correct."""
        # This would test that download functions have correct return type stubs
        # For now, document the requirement

        # Example: download functions should be typed as:
        # def download_to_file(response, output: None) -> bytes: ...
        # def download_to_file(response, output: Path) -> None: ...

        # This ensures type checkers understand the overloaded behavior
        pass

    def test_generic_type_parameters(self):
        """Test generic type parameters work correctly."""
        from kittycad.pagination import SyncPageIterator

        # Type should be parameterizable
        # iterator: SyncPageIterator[ApiCallWithPrice] = ...

        # This tests that the generic TypeVar T is properly used
        assert hasattr(SyncPageIterator, "__init__")


# Property-based testing with Hypothesis would go here
try:
    from hypothesis import given, strategies as st

    class TestUnsetHypothesis:
        """Property-based tests for Unset behavior using Hypothesis."""

        @given(st.text())
        def test_unset_vs_string_distinction(self, text_value: str):
            """Test UNSET is always distinguishable from string values."""
            assert UNSET is not text_value
            assert UNSET != text_value
            assert not (UNSET == text_value)

        @given(st.one_of(st.none(), st.text(), st.integers()))
        def test_unset_vs_other_values(self, value):
            """Test UNSET is distinguishable from other common values."""
            if value is not UNSET:  # Avoid the case where value is UNSET
                assert UNSET is not value
                assert UNSET != value

        def test_json_serialization_property(self):
            """Property-based test for JSON serialization with UNSET/None."""

            class TestModel(BaseModel):
                field: Union[str, None, Unset] = UNSET

                class Config:
                    exclude_unset = True

            # Test with UNSET
            model_unset = TestModel(field=UNSET)
            json_unset = model_unset.model_dump(exclude_unset=True)
            assert "field" not in json_unset

            # Test with None
            model_none = TestModel(field=None)
            json_none = model_none.model_dump(exclude_unset=True)
            assert "field" in json_none
            assert json_none["field"] is None

except ImportError:
    # Hypothesis not available, skip property-based tests
    pass


if __name__ == "__main__":
    pytest.main([__file__])



from typing import List, Optional

from ..models.text_to_cad_response import TextToCadResponse
from .base import KittyCadBaseModel


class TextToCadResponseResultsPage(KittyCadBaseModel):
    """A single page of results"""

    items: List[TextToCadResponse]

    next_page: Optional[str] = None



from enum import Enum


class GlobalAxis(str, Enum):
    """The global axes."""  # noqa: E501

    """# The X axis"""  # noqa: E501

    X = "x"

    """# The Y axis"""  # noqa: E501

    Y = "y"

    """# The Z axis"""  # noqa: E501

    Z = "z"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class UnitLength(str, Enum):
    """The valid types of length units."""  # noqa: E501

    """# Centimeters <https://en.wikipedia.org/wiki/Centimeter>"""  # noqa: E501

    CM = "cm"

    """# Feet <https://en.wikipedia.org/wiki/Foot_(unit)>"""  # noqa: E501

    FT = "ft"

    """# Inches <https://en.wikipedia.org/wiki/Inch>"""  # noqa: E501

    IN = "in"

    """# Meters <https://en.wikipedia.org/wiki/Meter>"""  # noqa: E501

    M = "m"

    """# Millimeters <https://en.wikipedia.org/wiki/Millimeter>"""  # noqa: E501

    MM = "mm"

    """# Yards <https://en.wikipedia.org/wiki/Yard>"""  # noqa: E501

    YD = "yd"

    def __str__(self) -> str:
        return str(self.value)



from typing import List, Optional

from ..models.api_call_with_price import ApiCallWithPrice
from .base import KittyCadBaseModel


class ApiCallWithPriceResultsPage(KittyCadBaseModel):
    """A single page of results"""

    items: List[ApiCallWithPrice]

    next_page: Optional[str] = None



from typing import Optional

from ..models.annotation_line_end_options import AnnotationLineEndOptions
from ..models.annotation_text_options import AnnotationTextOptions
from ..models.color import Color
from ..models.point3d import Point3d
from .base import KittyCadBaseModel


class AnnotationOptions(KittyCadBaseModel):
    """Options for annotations"""

    color: Optional[Color] = None

    line_ends: Optional[AnnotationLineEndOptions] = None

    line_width: Optional[float] = None

    position: Optional[Point3d] = None

    text: Optional[AnnotationTextOptions] = None



from enum import Enum


class OrgRole(str, Enum):
    """The roles in an organization."""  # noqa: E501

    """# Admins can do anything in the org."""  # noqa: E501

    ADMIN = "admin"

    """# Members of an org can not modify an org, but they belong in the org."""  # noqa: E501

    MEMBER = "member"

    """# A service account role."""  # noqa: E501

    SERVICE_ACCOUNT = "service_account"

    def __str__(self) -> str:
        return str(self.value)



#!/usr/bin/env python3
"""Test script to verify exception handling works correctly."""

from .. import KittyCAD
from ..exceptions import KittyCADAPIError, KittyCADClientError, KittyCADServerError


def test_successful_request():
    """Test that successful requests work without exceptions."""
    # Use a basic client that should work (assuming API is available)
    client = KittyCAD(token="dummy_token", base_url="https://httpbin.org")

    try:
        # This would fail with real API but let's test the exception structure
        result = client.meta.ping()
        assert result is not None or result is None  # Either outcome is acceptable
    except Exception:
        # Request failed as expected
        pass


def test_exception_types():
    """Test that our exception types are properly defined."""
    try:
        # Test base exception
        raise KittyCADAPIError("Test API error", 500, "TEST_ERROR", "req-123")
    except KittyCADAPIError as e:
        assert e.status_code == 500
        assert e.error_code == "TEST_ERROR"
        assert e.request_id == "req-123"

    try:
        # Test client error
        raise KittyCADClientError("Test client error", 404, "NOT_FOUND", "req-456")
    except KittyCADClientError as e:
        assert e.status_code == 404
        assert e.error_code == "NOT_FOUND"
        assert e.request_id == "req-456"

    try:
        # Test server error
        raise KittyCADServerError("Test server error", 500, "INTERNAL_ERROR", "req-789")
    except KittyCADServerError as e:
        assert e.status_code == 500
        assert e.error_code == "INTERNAL_ERROR"
        assert e.request_id == "req-789"


def test_exception_inheritance():
    """Test that exception inheritance works correctly."""
    try:
        raise KittyCADClientError("Test error", 400)
    except KittyCADAPIError as e:
        assert isinstance(e, KittyCADClientError)
        assert isinstance(e, KittyCADAPIError)

    try:
        raise KittyCADServerError("Test error", 500)
    except KittyCADAPIError as e:
        assert isinstance(e, KittyCADServerError)
        assert isinstance(e, KittyCADAPIError)


def test_import_structure():
    """Test that exceptions can be imported from the main package."""
    from kittycad import (
        KittyCADAPIError,
        KittyCADClientError,
        KittyCADConnectionError,
        KittyCADError,
        KittyCADServerError,
        KittyCADTimeoutError,
    )

    # Test that they are actually classes to avoid unused import warning
    exceptions = [
        KittyCADError,
        KittyCADAPIError,
        KittyCADClientError,
        KittyCADServerError,
        KittyCADConnectionError,
        KittyCADTimeoutError,
    ]
    assert all(isinstance(exc, type) for exc in exceptions)



from enum import Enum


class UnitEnergy(str, Enum):
    """The valid types of energy units."""  # noqa: E501

    """# British Thermal Unit (BTU) <https://en.wikipedia.org/wiki/British_thermal_unit>"""  # noqa: E501

    BTU = "btu"

    """# Electron Volts (eV) <https://en.wikipedia.org/wiki/Electronvolt>"""  # noqa: E501

    ELECTRONVOLTS = "electronvolts"

    """# Joules (or watt-seconds) <https://en.wikipedia.org/wiki/Joule>"""  # noqa: E501

    JOULES = "joules"

    """# Kilocalories (often just called calories) <https://en.wikipedia.org/wiki/Kilocalorie>"""  # noqa: E501

    KILOCALORIES = "kilocalories"

    """# Kilowatt hours (kWh) <https://en.wikipedia.org/wiki/Kilowatt-hour>"""  # noqa: E501

    KILOWATT_HOURS = "kilowatt_hours"

    """# Watt hours (Wh) <https://en.wikipedia.org/wiki/Kilowatt-hour>"""  # noqa: E501

    WATT_HOURS = "watt_hours"

    def __str__(self) -> str:
        return str(self.value)



from ..models.point3d import Point3d
from .base import KittyCadBaseModel


class GetSketchModePlane(KittyCadBaseModel):
    """The plane for sketch mode."""

    origin: Point3d

    x_axis: Point3d

    y_axis: Point3d

    z_axis: Point3d



from ..models.camera_settings import CameraSettings
from .base import KittyCadBaseModel


class ZoomToFit(KittyCadBaseModel):
    """The response from the `ZoomToFit` command."""

    settings: CameraSettings



from enum import Enum


class EntityType(str, Enum):
    """The type of entity"""  # noqa: E501

    ENTITY = "entity"

    OBJECT = "object"

    PATH = "path"

    CURVE = "curve"

    SOLID2D = "solid2d"

    SOLID3D = "solid3d"

    EDGE = "edge"

    FACE = "face"

    PLANE = "plane"

    VERTEX = "vertex"

    def __str__(self) -> str:
        return str(self.value)



from ..models.camera_view_state import CameraViewState
from .base import KittyCadBaseModel


class DefaultCameraGetView(KittyCadBaseModel):
    """The response from the `DefaultCameraGetView` command."""

    view: CameraViewState



from typing import List

from .base import KittyCadBaseModel


class PathGetCurveUuidsForVertices(KittyCadBaseModel):
    """The response from the `PathGetCurveUuidsForVertices` command."""

    curve_ids: List[str]



from enum import Enum


class AccountProvider(str, Enum):
    """An account provider."""  # noqa: E501

    """# The Apple account provider."""  # noqa: E501

    APPLE = "apple"

    """# The Discord account provider."""  # noqa: E501

    DISCORD = "discord"

    """# The Google account provider."""  # noqa: E501

    GOOGLE = "google"

    """# The GitHub account provider."""  # noqa: E501

    GITHUB = "github"

    """# The Microsoft account provider."""  # noqa: E501

    MICROSOFT = "microsoft"

    """# The SAML account provider."""  # noqa: E501

    SAML = "saml"

    """# The Tencent QQ account provider."""  # noqa: E501

    TENCENT = "tencent"

    def __str__(self) -> str:
        return str(self.value)



from typing import List

from .base import KittyCadBaseModel


class EntityGetSketchPaths(KittyCadBaseModel):
    """The response from the `EntityGetSketchPaths` command."""

    entity_ids: List[str]



from ..models.input_format3d import InputFormat3d
from ..models.output_format3d import OutputFormat3d
from .base import KittyCadBaseModel


class ConversionParams(KittyCadBaseModel):
    """Describes the file to convert (src) and what it should be converted into (output)."""

    output_format: OutputFormat3d

    src_format: InputFormat3d



from enum import Enum


class CurveType(str, Enum):
    """The type of Curve (embedded within path)"""  # noqa: E501

    LINE = "line"

    ARC = "arc"

    NURBS = "nurbs"

    def __str__(self) -> str:
        return str(self.value)



from ..models.camera_settings import CameraSettings
from .base import KittyCadBaseModel


class CameraDragMove(KittyCadBaseModel):
    """The response from the `CameraDragMove` command. Note this is an \"unreliable\" channel message, so this data may need more data like a \"sequence\" """

    settings: CameraSettings



from enum import Enum


class UnitVolume(str, Enum):
    """The valid types of volume units."""  # noqa: E501

    """# Cubic centimeters (cc or cm³) <https://en.wikipedia.org/wiki/Cubic_centimeter>"""  # noqa: E501

    CM3 = "cm3"

    """# Cubic feet (ft³) <https://en.wikipedia.org/wiki/Cubic_foot>"""  # noqa: E501

    FT3 = "ft3"

    """# Cubic inches (cu in or in³) <https://en.wikipedia.org/wiki/Cubic_inch>"""  # noqa: E501

    IN3 = "in3"

    """# Cubic meters (m³) <https://en.wikipedia.org/wiki/Cubic_meter>"""  # noqa: E501

    M3 = "m3"

    """# Cubic yards (yd³) <https://en.wikipedia.org/wiki/Cubic_yard>"""  # noqa: E501

    YD3 = "yd3"

    """# US Fluid Ounces (fl oz) <https://en.wikipedia.org/wiki/Fluid_ounce>"""  # noqa: E501

    USFLOZ = "usfloz"

    """# US Gallons (gal US) <https://en.wikipedia.org/wiki/Gallon>"""  # noqa: E501

    USGAL = "usgal"

    """# Liters (l) <https://en.wikipedia.org/wiki/Litre>"""  # noqa: E501

    L = "l"

    """# Milliliters (ml) <https://en.wikipedia.org/wiki/Litre>"""  # noqa: E501

    ML = "ml"

    def __str__(self) -> str:
        return str(self.value)
